from __future__ import annotations

import dataclasses
from dataclasses import replace
from typing import TYPE_CHECKING, Any, Collection, Generic, TypeVar
from warnings import warn

from typing_extensions import Annotated, TypeAlias, override

from litestar.dto.base_dto import AbstractDTO
from litestar.dto.data_structures import DTOFieldDefinition
from litestar.dto.field import DTO_FIELD_META_KEY, extract_dto_field
from litestar.exceptions import MissingDependencyException, ValidationException
from litestar.plugins.pydantic.utils import get_model_info, is_pydantic_2_model, is_pydantic_undefined, is_pydantic_v2
from litestar.types.empty import Empty
from litestar.typing import FieldDefinition

if TYPE_CHECKING:
    from typing import Generator

    from litestar.dto import DTOConfig

try:
    import pydantic as _  # noqa: F401
except ImportError as e:
    raise MissingDependencyException("pydantic") from e


try:
    import pydantic as pydantic_v2

    if not is_pydantic_v2(pydantic_v2):
        raise ImportError

    from pydantic import ValidationError as ValidationErrorV2
    from pydantic import v1 as pydantic_v1
    from pydantic.v1 import ValidationError as ValidationErrorV1

    ModelType: TypeAlias = "pydantic_v1.BaseModel | pydantic_v2.BaseModel"  # pyright: ignore[reportInvalidTypeForm,reportGeneralTypeIssues]

except ImportError:
    import pydantic as pydantic_v1  # type: ignore[no-redef]

    pydantic_v2 = Empty  # type: ignore[assignment]
    from pydantic import ValidationError as ValidationErrorV1  # type: ignore[assignment]

    ValidationErrorV2 = ValidationErrorV1  # type: ignore[assignment, misc]
    ModelType = "pydantic_v1.BaseModel"  # type: ignore[misc, assignment]


T = TypeVar("T", bound="ModelType | Collection[ModelType]")


__all__ = ("PydanticDTO",)

_down_types: dict[Any, Any] = {
    pydantic_v1.EmailStr: str,
    pydantic_v1.IPvAnyAddress: str,
    pydantic_v1.IPvAnyInterface: str,
    pydantic_v1.IPvAnyNetwork: str,
}

if pydantic_v2 is not Empty:  # type: ignore[comparison-overlap]  # pragma: no cover
    _down_types.update(
        {
            pydantic_v2.JsonValue: Any,
            pydantic_v2.EmailStr: str,
            pydantic_v2.IPvAnyAddress: str,
            pydantic_v2.IPvAnyInterface: str,
            pydantic_v2.IPvAnyNetwork: str,
            pydantic_v2.AwareDatetime: str,
        }
    )


def convert_validation_error(validation_error: ValidationErrorV1 | ValidationErrorV2) -> list[dict[str, Any]]:  # pyright: ignore[reportInvalidTypeForm,reportGeneralTypeIssues]
    error_list = validation_error.errors()
    for error in error_list:
        if isinstance(exception := error.get("ctx", {}).get("error"), Exception):
            error["ctx"]["error"] = type(exception).__name__  # pyright: ignore[reportTypedDictNotRequiredAccess]
    return error_list  # type: ignore[return-value]


def downtype_for_data_transfer(field_definition: FieldDefinition) -> FieldDefinition:
    if sub := _down_types.get(field_definition.annotation):
        return FieldDefinition.from_kwarg(
            annotation=Annotated[sub, field_definition.metadata], name=field_definition.name
        )
    return field_definition


class PydanticDTO(AbstractDTO[T], Generic[T]):
    """Support for domain modelling with Pydantic."""

    @override
    def decode_builtins(self, value: dict[str, Any]) -> Any:
        try:
            return super().decode_builtins(value)
        except (ValidationErrorV2, ValidationErrorV1) as ex:
            raise ValidationException(extra=convert_validation_error(ex)) from ex

    @override
    def decode_bytes(self, value: bytes) -> Any:
        try:
            return super().decode_bytes(value)
        except (ValidationErrorV2, ValidationErrorV1) as ex:
            raise ValidationException(extra=convert_validation_error(ex)) from ex

    @classmethod
    def generate_field_definitions(
        cls,
        model_type: type[pydantic_v1.BaseModel | pydantic_v2.BaseModel],  # pyright: ignore[reportInvalidTypeForm,reportGeneralTypeIssues]
    ) -> Generator[DTOFieldDefinition, None, None]:
        model_info = get_model_info(model_type)
        model_fields = model_info.model_fields
        model_field_definitions = model_info.field_definitions

        for field_name, field_definition in model_field_definitions.items():
            field_definition = downtype_for_data_transfer(field_definition)
            dto_field = extract_dto_field(field_definition, field_definition.extra)

            default: Any = Empty
            default_factory: Any = None
            if field_info := model_fields.get(field_name):
                # field_info might not exist, since FieldInfo isn't provided by pydantic
                # for computed fields, but we still generate a FieldDefinition for them
                try:
                    extra = field_info.extra  # type: ignore[union-attr]
                except AttributeError:
                    extra = field_info.json_schema_extra  # type: ignore[union-attr]

                if extra is not None and extra.pop(DTO_FIELD_META_KEY, None):
                    warn(
                        message="Declaring 'DTOField' via Pydantic's 'Field.extra' is deprecated. "
                        "Use 'Annotated', e.g., 'Annotated[str, DTOField(mark='read-only')]' instead. "
                        "Support for 'DTOField' in 'Field.extra' will be removed in v3.",
                        category=DeprecationWarning,
                        stacklevel=2,
                    )

                if not is_pydantic_undefined(field_info.default):
                    default = field_info.default
                elif field_definition.is_optional:
                    default = None
                else:
                    default = Empty

                default_factory = (
                    field_info.default_factory
                    if field_info.default_factory and not is_pydantic_undefined(field_info.default_factory)
                    else None
                )
            else:
                # Computed fields don't expose FieldInfo; mark Optional[...] as default=None so
                # the transfer struct treats them as optional when the computed value is None.
                if field_definition.is_optional:
                    default = None
                else:
                    default = Empty

            yield replace(
                DTOFieldDefinition.from_field_definition(
                    field_definition=field_definition,
                    dto_field=dto_field,
                    model_name=model_type.__name__,
                    default_factory=default_factory,
                    # we don't want the constraints to be set on the DTO struct as
                    # constraints, but as schema metadata only, so we can let pydantic
                    # handle all the constraining
                    passthrough_constraints=False,
                ),
                default=default,
                name=field_name,
            )

    @classmethod
    def detect_nested_field(cls, field_definition: FieldDefinition) -> bool:
        if pydantic_v2 is not Empty:  # type: ignore[comparison-overlap]
            return field_definition.is_subclass_of((pydantic_v1.BaseModel, pydantic_v2.BaseModel))
        return field_definition.is_subclass_of(pydantic_v1.BaseModel)  # type: ignore[unreachable]

    @classmethod
    def get_config_for_model_type(cls, config: DTOConfig, model_type: type[Any]) -> DTOConfig:
        if is_pydantic_2_model(model_type) and (model_config := getattr(model_type, "model_config", None)):
            if model_config.get("extra") == "forbid":
                config = dataclasses.replace(config, forbid_unknown_fields=True)
        elif issubclass(model_type, pydantic_v1.BaseModel) and (model_config := getattr(model_type, "Config", None)):  # noqa: SIM102
            if getattr(model_config, "extra", None) == "forbid":
                config = dataclasses.replace(config, forbid_unknown_fields=True)
        return config



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from litestar import Litestar, post
from litestar.plugins.sqlalchemy import SQLAlchemySerializationPlugin

if TYPE_CHECKING:
    from typing import List


class Base(DeclarativeBase): ...


class TodoItem(Base):
    __tablename__ = "todo_item"
    title: Mapped[str] = mapped_column(primary_key=True)
    done: Mapped[bool]


@post("/")
async def add_item(data: TodoItem) -> List[TodoItem]:
    return [data]


app = Litestar(route_handlers=[add_item], plugins=[SQLAlchemySerializationPlugin()])



from typing import Tuple, cast

import pytest
from _pytest.fixtures import FixtureRequest
from opentelemetry.metrics import get_meter_provider, set_meter_provider
from opentelemetry.sdk.metrics._internal import MeterProvider
from opentelemetry.sdk.metrics._internal.aggregation import (
    ExplicitBucketHistogramAggregation,
)
from opentelemetry.sdk.metrics._internal.export import InMemoryMetricReader
from opentelemetry.sdk.metrics._internal.instrument import Counter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import Span, TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor
from opentelemetry.sdk.trace.export.in_memory_span_exporter import InMemorySpanExporter

from litestar import WebSocket, get, websocket
from litestar.config.app import AppConfig
from litestar.contrib.opentelemetry import OpenTelemetryConfig, OpenTelemetryPlugin
from litestar.enums import ScopeType
from litestar.exceptions import http_exceptions
from litestar.status_codes import HTTP_200_OK
from litestar.testing import create_test_client
from litestar.types.asgi_types import ASGIApp, Receive, Scope, Send


@pytest.fixture(scope="session")
def resource() -> Resource:
    return Resource(attributes={SERVICE_NAME: "litestar-test"})


@pytest.fixture(scope="session")
def reader() -> InMemoryMetricReader:
    aggregation_last_value = {Counter: ExplicitBucketHistogramAggregation()}
    return InMemoryMetricReader(preferred_aggregation=aggregation_last_value)  # type: ignore[arg-type]


@pytest.fixture(scope="session")
def meter_provider(resource: Resource, reader: InMemoryMetricReader) -> MeterProvider:
    provider = MeterProvider(resource=resource, metric_readers=[reader])
    set_meter_provider(provider)
    return provider


@pytest.fixture()
def exporter() -> InMemorySpanExporter:
    return InMemorySpanExporter()


@pytest.fixture()
def config(
    resource: Resource, exporter: InMemorySpanExporter, meter_provider: MeterProvider, request: FixtureRequest
) -> OpenTelemetryConfig:
    tracer_provider = TracerProvider(resource=resource)
    tracer_provider.add_span_processor(SimpleSpanProcessor(exporter))
    meter = get_meter_provider().get_meter(f"litestar-test-{request.node.nodeid}")
    return OpenTelemetryConfig(tracer_provider=tracer_provider, meter=meter)


@pytest.fixture(params=["middleware", "plugin"])
def app_config(request: FixtureRequest, config: OpenTelemetryConfig) -> AppConfig:
    if request.param == "middleware":
        return AppConfig(middleware=[config.middleware])
    return AppConfig(plugins=[OpenTelemetryPlugin(config)])


def test_open_telemetry_middleware_with_http_route(
    app_config: AppConfig,
    reader: InMemoryMetricReader,
    exporter: InMemorySpanExporter,
) -> None:
    @get("/")
    def handler() -> dict:
        return {"hello": "world"}

    with create_test_client(handler, middleware=app_config.middleware, plugins=app_config.plugins) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert reader.get_metrics_data()

        first_span, second_span, third_span = cast("Tuple[Span, Span, Span]", exporter.get_finished_spans())
        assert dict(first_span.attributes) == {"http.status_code": 200, "asgi.event.type": "http.response.start"}  # type: ignore[arg-type]
        assert dict(second_span.attributes) == {"asgi.event.type": "http.response.body"}  # type: ignore[arg-type]
        assert dict(third_span.attributes) == {  # type: ignore[arg-type]
            "http.scheme": "http",
            "http.host": "testserver.local",
            "net.host.port": 80,
            "http.flavor": "1.1",
            "http.target": "/",
            "http.url": "http://testserver.local/",
            "http.method": "GET",
            "http.server_name": "testserver.local",
            "http.user_agent": "testclient",
            "net.peer.ip": "testclient",
            "net.peer.port": 50000,
            "http.route": "GET /",
            "http.status_code": 200,
        }

        metric_data = reader.get_metrics_data()
        assert metric_data
        assert metric_data.resource_metrics

        resource_metrics = metric_data.resource_metrics[0]
        assert resource_metrics.scope_metrics

        scope_metrics = resource_metrics.scope_metrics[0]
        assert scope_metrics.metrics

        request_metric = scope_metrics.metrics[0]
        assert len(list(request_metric.data.data_points)) == 1


def test_open_telemetry_middleware_with_websocket_route(
    app_config: AppConfig,
    reader: InMemoryMetricReader,
    exporter: InMemorySpanExporter,
) -> None:
    @websocket("/")
    async def handler(socket: "WebSocket") -> None:
        await socket.accept()
        await socket.send_json({"hello": "world"})
        await socket.close()

    with create_test_client(handler, middleware=app_config.middleware, plugins=app_config.plugins).websocket_connect(
        "/"
    ) as client:
        data = client.receive_json()
        assert data == {"hello": "world"}

        first_span, second_span, third_span, fourth_span, fifth_span = cast(
            "Tuple[Span, Span, Span, Span, Span]", exporter.get_finished_spans()
        )
        assert dict(first_span.attributes) == {"asgi.event.type": "websocket.connect"}  # type: ignore[arg-type]
        assert dict(second_span.attributes) == {"asgi.event.type": "websocket.accept"}  # type: ignore[arg-type]
        assert dict(third_span.attributes) == {"asgi.event.type": "websocket.send", "http.status_code": 200}  # type: ignore[arg-type]
        assert dict(fourth_span.attributes) == {"asgi.event.type": "websocket.close"}  # type: ignore[arg-type]
        assert dict(fifth_span.attributes) == {  # type: ignore[arg-type]
            "http.scheme": "ws",
            "http.host": "testserver.local",
            "net.host.port": 80,
            "http.target": "/",
            "http.url": "ws://testserver.local/",
            "http.server_name": "testserver.local",
            "http.user_agent": "testclient",
            "net.peer.ip": "testclient",
            "net.peer.port": 50000,
            "http.route": "/",
            "http.status_code": 200,
        }


def test_open_telemetry_middleware_handles_route_not_found_under_span_http(
    app_config: AppConfig,
    reader: InMemoryMetricReader,
    exporter: InMemorySpanExporter,
) -> None:
    @get("/")
    def handler() -> dict:
        raise Exception("random Exception")

    with create_test_client(handler, middleware=app_config.middleware, plugins=app_config.plugins) as client:
        response = client.get("/route_that_does_not_exist")
        assert response.status_code

        first_span, second_span, third_span = cast("Tuple[Span, Span, Span]", exporter.get_finished_spans())
        assert dict(first_span.attributes) == {  # type: ignore[arg-type]
            "http.status_code": 404,
            "asgi.event.type": "http.response.start",
        }
        assert dict(second_span.attributes) == {"asgi.event.type": "http.response.body"}  # type: ignore[arg-type]
        assert dict(third_span.attributes) == {  # type: ignore[arg-type]
            "http.scheme": "http",
            "http.host": "testserver.local",
            "net.host.port": 80,
            "http.flavor": "1.1",
            "http.target": "/route_that_does_not_exist",
            "http.url": "http://testserver.local/route_that_does_not_exist",
            "http.method": "GET",
            "http.server_name": "testserver.local",
            "http.user_agent": "testclient",
            "net.peer.ip": "testclient",
            "net.peer.port": 50000,
            "http.route": "GET /route_that_does_not_exist",
            "http.status_code": 404,
        }


def test_open_telemetry_middleware_handles_method_not_allowed_under_span_http(
    app_config: AppConfig,
    reader: InMemoryMetricReader,
    exporter: InMemorySpanExporter,
) -> None:
    @get("/")
    def handler() -> dict:
        raise Exception("random Exception")

    with create_test_client(handler, middleware=app_config.middleware, plugins=app_config.plugins) as client:
        response = client.post("/")
        assert response.status_code

        first_span, second_span, third_span = cast("Tuple[Span, Span, Span]", exporter.get_finished_spans())
        assert dict(first_span.attributes) == {  # type: ignore[arg-type]
            "http.status_code": 405,
            "asgi.event.type": "http.response.start",
        }
        assert dict(second_span.attributes) == {"asgi.event.type": "http.response.body"}  # type: ignore[arg-type]
        assert dict(third_span.attributes) == {  # type: ignore[arg-type]
            "http.scheme": "http",
            "http.host": "testserver.local",
            "net.host.port": 80,
            "http.flavor": "1.1",
            "http.target": "/",
            "http.url": "http://testserver.local/",
            "http.method": "POST",
            "http.server_name": "testserver.local",
            "http.user_agent": "testclient",
            "net.peer.ip": "testclient",
            "net.peer.port": 50000,
            "http.route": "POST /",
            "http.status_code": 405,
        }


def test_open_telemetry_middleware_handles_errors_caused_on_middleware(
    app_config: AppConfig,
    reader: InMemoryMetricReader,
    exporter: InMemorySpanExporter,
) -> None:
    raise_exception = True

    def middleware_factory(app: ASGIApp) -> ASGIApp:
        async def error_middleware(scope: Scope, receive: Receive, send: Send) -> None:
            if raise_exception:
                raise http_exceptions.NotAuthorizedException()
            await app(scope, receive, send)

        return error_middleware

    @get("/")
    def handler() -> dict:
        raise Exception("random Exception")

    with create_test_client(
        handler, middleware=[middleware_factory, *app_config.middleware], plugins=app_config.plugins
    ) as client:
        response = client.get("/")
        assert response.status_code

        first_span, second_span, third_span = cast("Tuple[Span, Span, Span]", exporter.get_finished_spans())
        assert dict(first_span.attributes) == {  # type: ignore[arg-type]
            "http.status_code": 401,
            "asgi.event.type": "http.response.start",
        }
        assert dict(second_span.attributes) == {"asgi.event.type": "http.response.body"}  # type: ignore[arg-type]
        assert dict(third_span.attributes) == {  # type: ignore[arg-type]
            "http.scheme": "http",
            "http.host": "testserver.local",
            "net.host.port": 80,
            "http.flavor": "1.1",
            "http.target": "/",
            "http.url": "http://testserver.local/",
            "http.method": "GET",
            "http.server_name": "testserver.local",
            "http.user_agent": "testclient",
            "net.peer.ip": "testclient",
            "net.peer.port": 50000,
            "http.route": "GET /",
            "http.status_code": 401,
        }


def test_after_exception_hook_handler_called_on_exception(
    resource: Resource, exporter: InMemorySpanExporter, meter_provider: MeterProvider, request: FixtureRequest
) -> None:
    """Test that after_exception_hook_handler is called when an exception occurs."""
    # Track calls to the hook
    hook_calls: list[tuple[Exception, Scope]] = []

    def after_exception_hook(exc: Exception, scope: Scope) -> None:
        hook_calls.append((exc, scope))

    # Create config with the hook handler
    tracer_provider = TracerProvider(resource=resource)
    tracer_provider.add_span_processor(SimpleSpanProcessor(exporter))
    meter = get_meter_provider().get_meter(f"litestar-test-{request.node.nodeid}")
    config = OpenTelemetryConfig(
        tracer_provider=tracer_provider, meter=meter, after_exception_hook_handler=after_exception_hook
    )

    test_exception = ValueError("test error")

    @get("/")
    def handler() -> dict:
        raise test_exception

    with create_test_client(handler, plugins=[OpenTelemetryPlugin(config)]) as client:
        response = client.get("/")
        assert response.status_code == 500

        # Verify the hook was called exactly once
        assert len(hook_calls) == 1
        exc, scope = hook_calls[0]

        # Verify the exception is the one we raised
        assert exc is test_exception
        assert isinstance(exc, ValueError)
        assert str(exc) == "test error"

        # Verify the scope is provided
        assert scope is not None
        assert scope["type"] == ScopeType.HTTP
        assert scope["path"] == "/"
        assert scope["method"] == "GET"


def test_after_exception_hook_handler_not_called_on_success(
    resource: Resource, exporter: InMemorySpanExporter, meter_provider: MeterProvider, request: FixtureRequest
) -> None:
    """Test that after_exception_hook_handler is not called when no exception occurs."""
    # Track calls to the hook
    hook_calls: list[tuple[Exception, Scope]] = []

    def after_exception_hook(exc: Exception, scope: Scope) -> None:
        hook_calls.append((exc, scope))

    # Create config with the hook handler
    tracer_provider = TracerProvider(resource=resource)
    tracer_provider.add_span_processor(SimpleSpanProcessor(exporter))
    meter = get_meter_provider().get_meter(f"litestar-test-{request.node.nodeid}")
    config = OpenTelemetryConfig(
        tracer_provider=tracer_provider, meter=meter, after_exception_hook_handler=after_exception_hook
    )

    @get("/")
    def handler() -> dict:
        return {"success": True}

    with create_test_client(handler, plugins=[OpenTelemetryPlugin(config)]) as client:
        response = client.get("/")
        assert response.status_code == 200

        # Verify the hook was NOT called
        assert len(hook_calls) == 0



from __future__ import annotations

from dataclasses import asdict, dataclass, field
from typing import Any

from litestar import MediaType, Request, Response
from litestar.exceptions import HTTPException, LitestarException
from litestar.exceptions.responses import _debug_response
from litestar.serialization import encode_json, get_serializer
from litestar.status_codes import HTTP_500_INTERNAL_SERVER_ERROR

__all__ = (
    "ExceptionResponseContent",
    "create_debug_response",
    "create_exception_response",
)


@dataclass
class ExceptionResponseContent:
    """Represent the contents of an exception-response."""

    status_code: int
    """Exception status code."""
    detail: str
    """Exception details or message."""
    media_type: MediaType | str
    """Media type of the response."""
    headers: dict[str, str] | None = field(default=None)
    """Headers to attach to the response."""
    extra: dict[str, Any] | list[Any] | None = field(default=None)
    """An extra mapping to attach to the exception."""

    def to_response(self, request: Request | None = None) -> Response:
        """Create a response from the model attributes.

        Returns:
            A response instance.
        """

        content: Any = {k: v for k, v in asdict(self).items() if k not in ("headers", "media_type") and v is not None}
        type_encoders = _debug_response._get_type_encoders_for_request(request) if request is not None else None

        if self.media_type != MediaType.JSON:
            content = encode_json(content, get_serializer(type_encoders))

        return Response(
            content=content,
            headers=self.headers,
            status_code=self.status_code,
            media_type=self.media_type,
            type_encoders=type_encoders,
        )


def create_exception_response(request: Request[Any, Any, Any], exc: Exception) -> Response:
    """Construct a response from an exception.

    Notes:
        - For instances of :class:`HTTPException <litestar.exceptions.HTTPException>` or other exception classes that have a
          ``status_code`` attribute (e.g. Starlette exceptions), the status code is drawn from the exception, otherwise
          response status is ``HTTP_500_INTERNAL_SERVER_ERROR``.

    Args:
        request: The request that triggered the exception.
        exc: An exception.

    Returns:
        Response: HTTP response constructed from exception details.
    """
    headers: dict[str, Any] | None
    extra: dict[str, Any] | list | None

    if isinstance(exc, HTTPException):
        status_code = exc.status_code
        headers = exc.headers
        extra = exc.extra
    else:
        status_code = HTTP_500_INTERNAL_SERVER_ERROR
        headers = None
        extra = None

    detail = (
        exc.detail
        if isinstance(exc, LitestarException) and status_code != HTTP_500_INTERNAL_SERVER_ERROR
        else "Internal Server Error"
    )

    try:
        media_type = request.route_handler.media_type
    except (KeyError, AttributeError):
        media_type = MediaType.JSON

    content = ExceptionResponseContent(
        status_code=status_code,
        detail=detail,
        headers=headers,
        extra=extra,
        media_type=media_type,
    )
    return content.to_response(request=request)


def create_debug_response(request: Request, exc: Exception) -> Response:
    """Create a debug response from an exception.

    Args:
        request: The request that triggered the exception.
        exc: An exception.

    Returns:
        Response: Debug response constructed from exception details.
    """
    return _debug_response.create_debug_response(request, exc)



from pydantic import BaseModel

from litestar import Litestar, Response, get
from litestar.datastructures import Cookie


class Resource(BaseModel):
    id: int
    name: str


@get("/resources", sync_to_thread=False)
def retrieve_resource() -> Response[Resource]:
    return Response(
        Resource(
            id=1,
            name="my resource",
        ),
        headers={"MY-HEADER": "xyz"},
        cookies=[Cookie(key="my-cookie", value="abc")],
    )


app = Litestar(route_handlers=[retrieve_resource])



from http import HTTPStatus

from litestar import Litestar, get
from litestar.config.cors import CORSConfig
from litestar.exceptions import HTTPException
from litestar.middleware import AbstractMiddleware
from litestar.testing import TestClient
from litestar.types.asgi_types import Receive, Scope, Send


class ExceptionMiddleware(AbstractMiddleware):
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        raise HTTPException(status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail="Intentional Error")


@get("/test")
async def handler() -> str:
    return "Should not reach this"


cors_config = CORSConfig(allow_methods=["GET"], allow_origins=["https://allowed-origin.com"], allow_credentials=True)
app = Litestar(route_handlers=[handler], cors_config=cors_config, middleware=[ExceptionMiddleware])


def test_cors_on_middleware_exception() -> None:
    with TestClient(app) as client:
        response = client.get("/test", headers={"Origin": "https://allowed-origin.com"})
        assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
        assert response.headers["access-control-allow-origin"] == "https://allowed-origin.com"
        assert response.headers["access-control-allow-credentials"] == "true"



import asyncio
import pdb  # noqa: T100
from concurrent.futures import ThreadPoolExecutor
from typing import Dict

# Install this packages if you want to run this test-app
import ipdb  # pyright: ignore  # noqa: T100
import pdbr  # pyright: ignore
import pudb  # pyright: ignore  # noqa: T100
import uvicorn

from litestar import Litestar, get


@get("/")
async def zero_division_error() -> Dict[str, str]:
    """Handler function that returns a greeting dictionary."""
    1 / 0  # pyright: ignore # noqa: B018
    return {"message": "ZeroDivisionError didn't occur."}


pdb_app = Litestar(route_handlers=[zero_division_error], pdb_on_exception=True, debugger_module=pdb)
ipdb_app = Litestar(route_handlers=[zero_division_error], pdb_on_exception=True, debugger_module=ipdb)
pudb_app = Litestar(route_handlers=[zero_division_error], pdb_on_exception=True, debugger_module=pudb)
pdbr_app = Litestar(route_handlers=[zero_division_error], pdb_on_exception=True, debugger_module=pdbr)


def run_server(app, port):
    uvicorn.run(app, port=port)


async def start_servers():
    with ThreadPoolExecutor() as executor:
        # Run all the servers concurrently
        await asyncio.gather(
            asyncio.get_event_loop().run_in_executor(executor, run_server, pdb_app, 8000),
            asyncio.get_event_loop().run_in_executor(executor, run_server, ipdb_app, 8001),
            asyncio.get_event_loop().run_in_executor(executor, run_server, pudb_app, 8002),
            asyncio.get_event_loop().run_in_executor(executor, run_server, pdbr_app, 8003),
        )


if __name__ == "__main__":
    asyncio.run(start_servers())



"""
Assemble components into an app that shall be tested
"""

import asyncio
from typing import AsyncIterator

from litestar import Litestar, get, post
from litestar.response import ServerSentEvent


@post("/log")
async def create_log_entry(message: str) -> None:
    # A print statement explicitly goes to buffer, logging may not
    print(f"Received log entry: {message}", flush=True)  # noqa: T201


@get("/notify/{topic:str}")
async def get_notified(topic: str) -> ServerSentEvent:
    async def generator() -> AsyncIterator[str]:
        yield topic
        while True:
            await asyncio.sleep(0.1)

    return ServerSentEvent(generator(), event_type="Notifier")


def create_test_app() -> Litestar:
    return Litestar(
        route_handlers=[create_log_entry, get_notified],
    )


app = create_test_app()



import asyncio

from litestar import Litestar, get
from litestar.config.response_cache import ResponseCacheConfig
from litestar.stores.redis import RedisStore


@get(cache=10)
async def something() -> str:
    await asyncio.sleep(1)
    return "something"


redis_store = RedisStore.with_client(url="redis://localhost/", port=6379, db=0)
cache_config = ResponseCacheConfig(store="redis_backed_store")
app = Litestar(
    [something],
    stores={"redis_backed_store": redis_store},
    response_cache_config=cache_config,
)



from __future__ import annotations

import dataclasses
import secrets
import sys
from dataclasses import asdict
from datetime import datetime, timedelta, timezone
from typing import Any, Sequence
from uuid import uuid4

import jwt
import pytest
from hypothesis import given
from hypothesis.strategies import datetimes

from litestar.exceptions import ImproperlyConfiguredException, NotAuthorizedException
from litestar.security.jwt import Token
from litestar.security.jwt.token import JWTDecodeOptions


@pytest.mark.parametrize("algorithm", ["HS256", "HS384", "HS512"])
@pytest.mark.parametrize("token_issuer", [None, "e3d7d10edbbc28bfebd8861d39ae7587acde1e1fcefe2cbbec686d235d68f475"])
@pytest.mark.parametrize("token_audience", [None, "627224198b4245ed91cf8353e4ccdf1650728c7ee92748f55fe1e9a9c4d961df"])
@pytest.mark.parametrize(
    "token_unique_jwt_id", [None, "10f5c6967783ddd6bb0c4e8262d7097caeae64705e45f83275e3c32eee5d30f2"]
)
@pytest.mark.parametrize("token_extras", [None, {"email": "test@test.com"}])
def test_token(
    algorithm: str,
    token_issuer: str | None,
    token_audience: str | None,
    token_unique_jwt_id: str | None,
    token_extras: dict[str, Any] | None,
) -> None:
    token_secret = secrets.token_hex()
    token = Token(
        sub=secrets.token_hex(),
        exp=(datetime.now(timezone.utc) + timedelta(minutes=30)),
        aud=token_audience,
        iss=token_issuer,
        jti=token_unique_jwt_id,
        extras=token_extras or {},
    )
    encoded_token = token.encode(secret=token_secret, algorithm=algorithm)
    decoded_token = token.decode(encoded_token=encoded_token, secret=token_secret, algorithm=algorithm)
    assert asdict(token) == asdict(decoded_token)


@pytest.mark.parametrize(
    "algorithm, secret",
    [
        (
            "nope",
            "1",
        ),
        (
            "",
            None,
        ),
        (
            "",
            "",
        ),
        (
            "",
            "1",
        ),
    ],
)
def test_encode_validation(algorithm: str, secret: str) -> None:
    with pytest.raises(ImproperlyConfiguredException):
        Token(
            sub="123",
            exp=(datetime.now(timezone.utc) + timedelta(seconds=30)),
        ).encode(algorithm=algorithm, secret=secret)


def test_decode_validation() -> None:
    token = Token(
        sub="123",
        exp=(datetime.now(timezone.utc) + timedelta(seconds=30)),
    )
    algorithm = "HS256"
    secret = uuid4().hex
    encoded_token = token.encode(algorithm=algorithm, secret=secret)

    token.decode(encoded_token=encoded_token, algorithm=algorithm, secret=secret)

    with pytest.raises(NotAuthorizedException):
        token.decode(encoded_token=secret, algorithm=algorithm, secret=secret)

    with pytest.raises(NotAuthorizedException):
        token.decode(encoded_token=encoded_token, algorithm="nope", secret=secret)

    with pytest.raises(NotAuthorizedException):
        token.decode(encoded_token=encoded_token, algorithm=algorithm, secret=uuid4().hex)


@given(exp=datetimes(min_value=datetime(1970, 1, 1), max_value=datetime.now() - timedelta(seconds=1)))
def test_exp_validation(exp: datetime) -> None:
    if sys.platform == "win32" and exp == datetime(1970, 1, 1):
        # this does not work on windows. see https://bugs.python.org/issue29097
        pytest.skip("Skipping because .timestamp is weird on windows sometimes")

    with pytest.raises(ImproperlyConfiguredException):
        Token(
            sub="123",
            exp=exp,
            iat=(datetime.now() - timedelta(seconds=30)),
        )


@given(iat=datetimes(min_value=datetime.now() + timedelta(days=1)))
def test_iat_validation(iat: datetime) -> None:
    if sys.platform == "win32" and iat >= datetime(3000, 1, 1, 0, 0):
        # this does not work on windows. see https://bugs.python.org/issue29097
        pytest.skip("Skipping because .timestamp is weird on windows sometimes")

    with pytest.raises(ImproperlyConfiguredException):
        Token(
            sub="123",
            iat=iat,
            exp=(iat + timedelta(seconds=120)),
        )


def test_sub_validation() -> None:
    with pytest.raises(ImproperlyConfiguredException):
        Token(
            sub="",
            iat=(datetime.now() - timedelta(seconds=30)),
            exp=(datetime.now() + timedelta(seconds=120)),
        )


def test_extra_fields() -> None:
    raw_token = {
        "sub": secrets.token_hex(),
        "iat": datetime.now(timezone.utc),
        "azp": "extra value",
        "email": "thetest@test.com",
        "exp": (datetime.now(timezone.utc) + timedelta(seconds=30)),
    }
    token_secret = secrets.token_hex()
    encoded_token = jwt.encode(payload=raw_token, key=token_secret, algorithm="HS256")
    token = Token.decode(encoded_token=encoded_token, secret=token_secret, algorithm="HS256")
    assert "azp" in token.extras
    assert "email" in token.extras

    raw_token = {
        "sub": secrets.token_hex(),
        "iat": datetime.now(timezone.utc),
        "exp": (datetime.now(timezone.utc) + timedelta(seconds=30)),
    }
    token_secret = secrets.token_hex()
    encoded_token = jwt.encode(payload=raw_token, key=token_secret, algorithm="HS256")
    token = Token.decode(encoded_token=encoded_token, secret=token_secret, algorithm="HS256")
    assert token.extras == {}


@pytest.mark.parametrize("audience", [None, ["foo", "bar"]])
def test_strict_aud_with_multiple_audiences_raises(audience: str | list[str]) -> None:
    with pytest.raises(ValueError, match="When using 'strict_audience=True'"):
        Token.decode(
            "",
            secret="",
            algorithm="HS256",
            audience=audience,
            strict_audience=True,
        )


@pytest.mark.parametrize("audience", ["foo", ["foo"]])
def test_strict_aud_with_one_element_sequence(audience: str | list[str]) -> None:
    # when validating with strict audience, PyJWT requires that the 'audience' parameter
    # is passed as a string - one element lists are not allowed. Since we allow these
    # generally, we convert them to a string in this case
    secret = secrets.token_hex()
    encoded = Token(exp=datetime.now() + timedelta(days=1), sub="foo", aud="foo").encode(secret, "HS256")
    Token.decode(
        encoded,
        secret=secret,
        algorithm="HS256",
        audience=audience,
        strict_audience=True,
    )


@pytest.mark.parametrize(
    "audience",
    [
        pytest.param(None, id="None"),
        pytest.param("foo", id="String"),
        pytest.param("not-foo", id="InvalidAudience"),
        pytest.param(["foo", "bar"], id="List"),
    ],
)
def test_validate_audience(audience: str | list[str]) -> None:
    secret = secrets.token_hex()
    encoded = Token(exp=datetime.now() + timedelta(days=1), sub="foo", aud=["foo", "bar"]).encode(secret, "HS256")

    def decode() -> None:
        Token.decode(
            encoded,
            secret=secret,
            algorithm="HS256",
            audience=audience,
        )

    if audience != "not-foo":
        decode()
    else:
        with pytest.raises(NotAuthorizedException):
            decode()


def test_custom_decode_payload() -> None:
    @dataclasses.dataclass
    class CustomToken(Token):
        @classmethod
        def decode_payload(
            cls,
            encoded_token: str,
            secret: str,
            algorithms: list[str],
            issuer: str | Sequence[str] | None = None,
            audience: str | Sequence[str] | None = None,
            options: JWTDecodeOptions | None = None,
        ) -> Any:
            payload = super().decode_payload(
                encoded_token=encoded_token,
                secret=secret,
                algorithms=algorithms,
            )
            payload["sub"] = "some-random-value"
            return payload

    _secret = secrets.token_hex()
    encoded = CustomToken(exp=datetime.now() + timedelta(days=1), sub="foo").encode(_secret, "HS256")
    assert CustomToken.decode(encoded, secret=_secret, algorithm="HS256").sub == "some-random-value"


@pytest.mark.parametrize("issuer", [None, "text", ["list", "of", "values"]])
def test_token_issuer(issuer: str | list[str] | None) -> None:
    iss = issuer[0] if isinstance(issuer, list) else issuer
    secret = secrets.token_hex()
    encoded = Token(
        exp=datetime.now() + timedelta(days=1),
        sub="foo",
        iss=iss,
    ).encode(secret, "HS256")

    token = Token.decode(
        encoded,
        secret=secret,
        algorithm="HS256",
        issuer=issuer,
    )

    assert token.iss == iss



from __future__ import annotations

import dataclasses
from dataclasses import asdict, dataclass, field
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence, TypedDict, Union

import jwt
import msgspec

from litestar.exceptions import ImproperlyConfiguredException, NotAuthorizedException

if TYPE_CHECKING:
    from typing_extensions import Self

__all__ = (
    "JWTDecodeOptions",
    "Token",
)


def _normalize_datetime(value: datetime) -> datetime:
    """Convert the given value into UTC and strip microseconds.

    Args:
        value: A datetime instance

    Returns:
        A datetime instance
    """
    if value.tzinfo is not None:
        value = value.astimezone(timezone.utc)

    return value.replace(microsecond=0)


class JWTDecodeOptions(TypedDict, total=False):
    """``options`` for PyJWTs :func:`jwt.decode`"""

    verify_aud: bool
    verify_iss: bool
    verify_exp: bool
    verify_nbf: bool
    strict_aud: bool
    require: list[str]


@dataclass
class Token:
    """JWT Token DTO."""

    exp: datetime
    """Expiration - datetime for token expiration."""
    sub: str
    """Subject - usually a unique identifier of the user or equivalent entity."""
    iat: datetime = field(default_factory=lambda: _normalize_datetime(datetime.now(timezone.utc)))
    """Issued at - should always be current now."""
    iss: Optional[str] = field(default=None)  # noqa: UP045
    """Issuer - optional unique identifier for the issuer."""
    aud: Optional[Union[str, Sequence[str]]] = field(default=None)  # noqa: UP045, UP007
    """Audience - intended audience."""
    jti: Optional[Union[str, Sequence[str]]] = field(default=None)  # noqa: UP045, UP007
    """JWT ID - a unique identifier of the JWT between different issuers."""
    extras: Dict[str, Any] = field(default_factory=dict)  # noqa: UP006
    """Extra fields that were found on the JWT token."""

    def __post_init__(self) -> None:
        if len(self.sub) < 1:
            raise ImproperlyConfiguredException("sub must be a string with a length greater than 0")

        if isinstance(self.exp, datetime) and (
            (exp := _normalize_datetime(self.exp)).timestamp()
            >= _normalize_datetime(datetime.now(timezone.utc)).timestamp()
        ):
            self.exp = exp
        else:
            raise ImproperlyConfiguredException("exp value must be a datetime in the future")

        if isinstance(self.iat, datetime) and (
            (iat := _normalize_datetime(self.iat)).timestamp()
            <= _normalize_datetime(datetime.now(timezone.utc)).timestamp()
        ):
            self.iat = iat
        else:
            raise ImproperlyConfiguredException("iat must be a current or past time")

    @classmethod
    def decode_payload(
        cls,
        encoded_token: str,
        secret: str,
        algorithms: list[str],
        issuer: str | Sequence[str] | None = None,
        audience: str | Sequence[str] | None = None,
        options: JWTDecodeOptions | None = None,
    ) -> Any:
        """Decode and verify the JWT and return its payload"""
        return jwt.decode(
            jwt=encoded_token,
            key=secret,
            algorithms=algorithms,
            issuer=issuer,  # type: ignore[arg-type]
            audience=audience,
            options=options,  # type: ignore[arg-type]
        )

    @classmethod
    def decode(
        cls,
        encoded_token: str,
        secret: str,
        algorithm: str,
        audience: str | Sequence[str] | None = None,
        issuer: str | Sequence[str] | None = None,
        require_claims: Sequence[str] | None = None,
        verify_exp: bool = True,
        verify_nbf: bool = True,
        strict_audience: bool = False,
    ) -> Self:
        """Decode a passed in token string and return a Token instance.

        Args:
            encoded_token: A base64 string containing an encoded JWT.
            secret: The secret with which the JWT is encoded.
            algorithm: The algorithm used to encode the JWT.
            audience: Verify the audience when decoding the token. If the audience in
                the token does not match any audience given, raise a
                :exc:`NotAuthorizedException`
            issuer: Verify the issuer when decoding the token. If the issuer in the
                token does not match any issuer given, raise a
                :exc:`NotAuthorizedException`
            require_claims: Verify that the given claims are present in the token
            verify_exp: Verify that the value of the ``exp`` (*expiration*) claim is in
                the future
            verify_nbf: Verify that the value of the ``nbf`` (*not before*) claim is in
                the past
            strict_audience: Verify that the value of the ``aud`` (*audience*) claim is
                a single value, and not a list of values, and matches ``audience``
                exactly. Requires the value passed to the ``audience`` to be a sequence
                of length 1

        Returns:
            A decoded Token instance.

        Raises:
            NotAuthorizedException: If the token is invalid.
        """

        options: JWTDecodeOptions = {
            "verify_aud": bool(audience),
            "verify_iss": bool(issuer),
        }
        if require_claims:
            options["require"] = list(require_claims)
        if verify_exp is False:
            options["verify_exp"] = False
        if verify_nbf is False:
            options["verify_nbf"] = False
        if strict_audience:
            if audience is None or (not isinstance(audience, str) and len(audience) != 1):
                raise ValueError("When using 'strict_audience=True', 'audience' must be a sequence of length 1")
            options["strict_aud"] = True
            # although not documented, pyjwt requires audience to be a string if
            # using the strict_aud option
            if not isinstance(audience, str):
                audience = audience[0]

        try:
            payload = cls.decode_payload(
                encoded_token=encoded_token,
                secret=secret,
                algorithms=[algorithm],
                audience=audience,
                issuer=issuer,
                options=options,
            )
            # msgspec can do these conversions as well, but to keep backwards
            # compatibility, we do it ourselves, since the datetime parsing works a
            # little bit different there
            payload["exp"] = datetime.fromtimestamp(payload["exp"], tz=timezone.utc)
            payload["iat"] = datetime.fromtimestamp(payload["iat"], tz=timezone.utc)
            extra_fields = payload.keys() - {f.name for f in dataclasses.fields(cls)}
            extras = payload.setdefault("extras", {})
            for key in extra_fields:
                extras[key] = payload.pop(key)
            return msgspec.convert(payload, cls, strict=False)
        except (
            KeyError,
            jwt.exceptions.InvalidTokenError,
            ImproperlyConfiguredException,
            msgspec.ValidationError,
        ) as e:
            raise NotAuthorizedException("Invalid token") from e

    def encode(self, secret: str, algorithm: str) -> str:
        """Encode the token instance into a string.

        Args:
            secret: The secret with which the JWT is encoded.
            algorithm: The algorithm used to encode the JWT.

        Returns:
            An encoded token string.

        Raises:
            ImproperlyConfiguredException: If encoding fails.
        """
        try:
            return jwt.encode(
                payload={k: v for k, v in asdict(self).items() if v is not None},
                key=secret,
                algorithm=algorithm,
            )
        except (jwt.DecodeError, NotImplementedError) as e:
            raise ImproperlyConfiguredException("Failed to encode token") from e



from __future__ import annotations

import pytest

from litestar.types.empty import Empty
from litestar.utils.empty import EmptyValueError, value_or_raise


def test_value_or_raise_empty() -> None:
    with pytest.raises(EmptyValueError):
        value_or_raise(Empty)


def test_value_or_raise_value() -> None:
    assert value_or_raise(1) == 1



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import select

from litestar import Litestar, post
from litestar.plugins.sqlalchemy import SQLAlchemyInitPlugin, SQLAlchemySyncConfig

if TYPE_CHECKING:
    from typing import Tuple

    from sqlalchemy import Engine
    from sqlalchemy.orm import Session


@post("/", sync_to_thread=True)
def handler(db_session: Session, db_engine: Engine) -> Tuple[int, int]:
    one = db_session.execute(select(1)).scalar()

    with db_engine.begin() as conn:
        two = conn.execute(select(2)).scalar()

    return one, two


config = SQLAlchemySyncConfig(connection_string="sqlite:///sync.sqlite")
plugin = SQLAlchemyInitPlugin(config=config)
app = Litestar(route_handlers=[handler], plugins=[plugin])



"""Litestar CLI."""

from __future__ import annotations

from importlib.util import find_spec

if find_spec("rich_click") is not None:  # pragma: no cover
    import rich_click

    rich_click.rich_click.TEXT_MARKUP = "rich"
    rich_click.rich_click.SHOW_ARGUMENTS = True
    rich_click.rich_click.GROUP_ARGUMENTS_OPTIONS = True
    rich_click.rich_click.STYLE_ERRORS_SUGGESTION = "magenta italic"
    rich_click.rich_click.ERRORS_SUGGESTION = ""
    rich_click.rich_click.ERRORS_EPILOGUE = ""
    rich_click.rich_click.MAX_WIDTH = 120
    rich_click.rich_click.THEME = "star"


from .main import litestar_group

__all__ = ["litestar_group"]



from typing import Any, Dict

import pytest
from prometheus_client import REGISTRY

from litestar import Controller, Litestar, Request, get
from litestar.plugins.prometheus import PrometheusMiddleware
from litestar.status_codes import HTTP_200_OK
from litestar.testing import TestClient


def clear_collectors() -> None:
    collectors = list(REGISTRY._collector_to_names.keys())
    for collector in collectors:
        REGISTRY.unregister(collector)

    PrometheusMiddleware._metrics = {}


@pytest.mark.parametrize(
    "group_path, route_path, route_template, expected_path",
    [
        (True, "/test/litestar", "test/litestar", "/test/litestar"),
        (False, "/test/litestar", "test/litestar", "/test/litestar"),
        (True, "/test/litestar", "test/{name:str}", "/test/{name}"),
        (False, "/test/litestar", "test/{name:str}", "/test/litestar"),
        (
            True,
            "/project/123a/team/abc/test/hi",
            "project/{project:str}/team/{team:str}/test/{name:str}",
            "/project/{project}/team/{team}/test/{name}",
        ),
        (
            False,
            "/project/123a/team/abc/test/hi",
            "project/{project:str}/team/{team:str}/test/{name:str}",
            "/project/123a/team/abc/test/hi",
        ),
    ],
)
def test_prometheus_exporter_example(
    group_path: bool, route_path: str, route_template: str, expected_path: str
) -> None:
    from docs.examples.plugins.prometheus.using_prometheus_exporter import create_app

    app = create_app(group_path=group_path)

    clear_collectors()

    @get(route_template)
    def home(name: str) -> Dict[str, Any]:
        return {"hello": name}

    app.register(home)

    with TestClient(app) as client:
        client.get(route_path)
        metrics_exporter_response = client.get("/metrics")

        assert metrics_exporter_response.status_code == HTTP_200_OK
        metrics = metrics_exporter_response.content.decode()
        assert expected_path in metrics


def test_correct_population_path_template() -> None:
    class TestController(Controller):
        path = "/prefix"

        @get("/{id_:int}")
        async def b(self, request: Request, id_: int) -> str:
            return request.scope["path_template"]

        @get("/{id_:int}/postfix")
        async def a(self, request: Request, id_: int) -> str:
            return request.scope["path_template"]

    app = Litestar([TestController])

    with TestClient(app) as client:
        without_postfix_resp = client.get("/prefix/1")
        with_postfix_resp = client.get("/prefix/1/postfix")

        assert without_postfix_resp.content.decode() == "/prefix/{id_}"
        assert with_postfix_resp.content.decode() == "/prefix/{id_}/postfix"



from http import HTTPStatus

from litestar import Litestar, route
from litestar.config.cors import CORSConfig
from litestar.enums import HttpMethod
from litestar.response import Response
from litestar.testing import TestClient


@route("/custom-options", http_method=HttpMethod.OPTIONS)
async def custom_options_handler() -> Response[str]:
    return Response(
        status_code=200,
        headers={"Custom-Handler": "Active"},
        content="Handled by Custom Options",
    )


cors_config = CORSConfig(allow_methods=["GET", "OPTIONS"], allow_origins=["https://allowed-origin.com"])
app = Litestar(route_handlers=[custom_options_handler], cors_config=cors_config)


def test_custom_options_handler_cors_pre_flight_request() -> None:
    with TestClient(app) as client:
        response = client.options(
            "/custom-options", headers={"Origin": "https://allowed-origin.com", "Access-Control-Request-Method": "GET"}
        )
        assert response.status_code == HTTPStatus.NO_CONTENT
        assert "access-control-allow-origin" in response.headers
        assert "Custom-Handler" not in response.headers


def test_custom_options_handler_non_cors_request() -> None:
    with TestClient(app) as client:
        response = client.options("/custom-options")
        assert response.status_code == 200
        assert response.headers.get("Custom-Handler") == "Active"
        assert response.text == "Handled by Custom Options"



import pytest

from litestar import Controller, Litestar, Router, post
from litestar.exceptions import ImproperlyConfiguredException
from litestar.types import Empty


def test_resolve_request_max_body_size() -> None:
    @post("/1")
    def router_handler() -> None:
        pass

    @post("/2")
    def app_handler() -> None:
        pass

    class MyController(Controller):
        request_max_body_size = 2

        @post("/3")
        def controller_handler(self) -> None:
            pass

    router = Router("/", route_handlers=[router_handler], request_max_body_size=1)
    app = Litestar(route_handlers=[app_handler, router, MyController], request_max_body_size=3)
    assert router_handler.resolve_request_max_body_size() == 1
    assert app_handler.resolve_request_max_body_size() == 3
    assert (
        next(r for r in app.routes if r.path == "/3").route_handler_map["POST"][0].resolve_request_max_body_size() == 2  # type: ignore[union-attr]
    )


def test_resolve_request_max_body_size_none() -> None:
    @post("/1", request_max_body_size=None)
    def router_handler() -> None:
        pass

    Litestar([router_handler])
    assert router_handler.resolve_request_max_body_size() is None


def test_resolve_request_max_body_size_app_default() -> None:
    @post("/")
    def router_handler() -> None:
        pass

    app = Litestar(route_handlers=[router_handler])

    assert router_handler.resolve_request_max_body_size() == app.request_max_body_size == 10_000_000


def test_resolve_request_max_body_size_empty_on_all_layers_raises() -> None:
    @post("/")
    def handler_one() -> None:
        pass

    Litestar([handler_one], request_max_body_size=Empty)  # type: ignore[arg-type]
    with pytest.raises(ImproperlyConfiguredException):
        handler_one.resolve_request_max_body_size()

    @post("/")
    def handler_two() -> None:
        pass

    with pytest.raises(ImproperlyConfiguredException):
        handler_two.resolve_request_max_body_size()



from __future__ import annotations

from dataclasses import dataclass

from litestar import Litestar, get
from litestar.dto import DataclassDTO, DTOConfig


@dataclass
class Person:
    name: str
    age: int
    email: str


class ReadDTO(DataclassDTO[Person]):
    config = DTOConfig(exclude={"email"})


@get("/person/{name:str}", return_dto=ReadDTO, sync_to_thread=False)
def get_person(name: str) -> Person:
    # Your logic to retrieve the person goes here
    # For demonstration purposes, a placeholder Person instance is returned
    return Person(name=name, age=30, email=f"email_of_{name}@example.com")


app = Litestar(route_handlers=[get_person])



from dataclasses import dataclass
from typing import List

from litestar import Litestar, get


@dataclass
class TodoItem:
    title: str
    done: bool


TODO_LIST: List[TodoItem] = [
    TodoItem(title="Start writing TODO list", done=True),
    TodoItem(title="???", done=False),
    TodoItem(title="Profit", done=False),
]


@get("/")
async def get_list(done: str) -> List[TodoItem]:
    if done == "1":
        return [item for item in TODO_LIST if item.done]
    return [item for item in TODO_LIST if not item.done]


app = Litestar([get_list])



from typing import TYPE_CHECKING, List, cast

from sqlalchemy import func, select
from sqlalchemy.orm import Mapped

from litestar import Litestar, get
from litestar.di import Provide
from litestar.pagination import AbstractAsyncOffsetPaginator, OffsetPagination
from litestar.plugins.sqlalchemy import SQLAlchemyAsyncConfig, SQLAlchemyInitPlugin, base

if TYPE_CHECKING:
    from sqlalchemy.engine.result import ScalarResult
    from sqlalchemy.ext.asyncio import AsyncSession


class Person(base.UUIDBase):
    name: Mapped[str]


class PersonOffsetPaginator(AbstractAsyncOffsetPaginator[Person]):
    def __init__(self, async_session: AsyncSession) -> None:  # 'async_session' dependency will be injected here.
        self.async_session = async_session

    async def get_total(self) -> int:
        return cast("int", await self.async_session.scalar(select(func.count(Person.id))))

    async def get_items(self, limit: int, offset: int) -> List[Person]:
        people: ScalarResult = await self.async_session.scalars(select(Person).slice(offset, limit))
        return list(people.all())


# Create a route handler. The handler will receive two query parameters - 'limit' and 'offset', which is passed
# to the paginator instance. Also create a dependency 'paginator' which will be injected into the handler.
@get("/people", dependencies={"paginator": Provide(PersonOffsetPaginator)})
async def people_handler(paginator: PersonOffsetPaginator, limit: int, offset: int) -> OffsetPagination[Person]:
    return await paginator(limit=limit, offset=offset)


sqlalchemy_config = SQLAlchemyAsyncConfig(
    connection_string="sqlite+aiosqlite:///test.sqlite", session_dependency_key="async_session"
)  # Create 'async_session' dependency.
sqlalchemy_plugin = SQLAlchemyInitPlugin(config=sqlalchemy_config)


async def on_startup() -> None:
    """Initializes the database."""
    async with sqlalchemy_config.get_engine().begin() as conn:
        await conn.run_sync(base.UUIDBase.metadata.create_all)


app = Litestar(route_handlers=[people_handler], on_startup=[on_startup], plugins=[sqlalchemy_plugin])



from typing import TYPE_CHECKING

from starlette.applications import Starlette
from starlette.responses import JSONResponse
from starlette.routing import Route

from litestar import Litestar, asgi

if TYPE_CHECKING:
    from starlette.requests import Request


async def index(request: "Request") -> JSONResponse:
    """A generic starlette handler."""
    return JSONResponse({"forwarded_path": request.url.path})


starlette_app = asgi(path="/some/sub-path", is_mount=True, copy_scope=True)(
    Starlette(
        routes=[
            Route("/", index),
            Route("/abc/", index),
            Route("/123/another/sub-path/", index),
        ],
    )
)


app = Litestar(route_handlers=[starlette_app])



from litestar import Litestar, WebSocket, websocket
from litestar.channels import ChannelsPlugin
from litestar.channels.backends.memory import MemoryChannelsBackend


@websocket("/ws")
async def handler(socket: WebSocket, channels: ChannelsPlugin) -> None:
    await socket.accept()

    async with channels.start_subscription(["some_channel"]) as subscriber:
        await channels.put_subscriber_history(subscriber, ["some_channel"], limit=10)


app = Litestar(
    [handler],
    plugins=[ChannelsPlugin(backend=MemoryChannelsBackend(history=20))],
)



from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator

import anyio

from litestar import Litestar, WebSocket, websocket_listener
from litestar.exceptions import WebSocketDisconnect
from litestar.handlers import send_websocket_stream


@asynccontextmanager
async def listener_lifespan(socket: WebSocket) -> AsyncGenerator[None, Any]:
    is_closed = anyio.Event()

    async def handle_stream() -> AsyncGenerator[str, None]:
        while not is_closed.is_set():
            await anyio.sleep(0.1)
            yield "ping"

    async with anyio.create_task_group() as tg:
        tg.start_soon(send_websocket_stream, socket, handle_stream())

        try:
            yield
        except WebSocketDisconnect:
            pass
        finally:
            is_closed.set()


@websocket_listener("/", connection_lifespan=listener_lifespan)
def handler(data: str) -> str:
    return data


app = Litestar([handler])



from pathlib import Path

from litestar import Litestar, get
from litestar.contrib.minijinja import MiniJinjaTemplateEngine, StateProtocol
from litestar.response import Template
from litestar.template.config import TemplateConfig


def my_template_function(ctx: StateProtocol) -> str:
    return ctx.lookup("my_context_key") or "nope"


def register_template_callables(engine: MiniJinjaTemplateEngine) -> None:
    engine.register_template_callable(key="check_context_key", template_callable=my_template_function)


template_config = TemplateConfig(
    directory=Path(__file__).parent / "templates",
    engine=MiniJinjaTemplateEngine,
    engine_callback=register_template_callables,
)


@get("/", sync_to_thread=False)
def index() -> Template:
    return Template(template_name="index.html.minijinja")


app = Litestar(route_handlers=[index], template_config=template_config)



from litestar import Litestar, WebSocket
from litestar.handlers import WebsocketListener


class Handler(WebsocketListener):
    path = "/"

    def on_accept(self, socket: WebSocket) -> None:
        print("Connection accepted")

    def on_disconnect(self, socket: WebSocket) -> None:
        print("Connection closed")

    def on_receive(self, data: str) -> str:
        return data


app = Litestar([Handler])



from litestar import Litestar, WebSocket, websocket_listener


async def accept_connection(socket: WebSocket) -> None:
    await socket.accept(headers={"Cookie": "custom-cookie"})


@websocket_listener("/", connection_accept_handler=accept_connection)
def handler(data: str) -> str:
    return data


app = Litestar([handler])



from __future__ import annotations

from typing import Any, Awaitable, Callable, Dict

from typing_extensions import TypeAlias

from litestar.connection import ASGIConnection

Extractor: TypeAlias = Callable[[Dict[str, Any], ASGIConnection], Awaitable[None]]



from __future__ import annotations

from typing import TYPE_CHECKING

from litestar import Controller, post

if TYPE_CHECKING:
    from .domain import Model


class MyController(Controller):
    @post(sync_to_thread=False)
    def post_handler(self, data: Model) -> Model:
        return data



from typing import Dict, Generator

from litestar import Litestar, get
from litestar.di import Provide

STATE = {"result": None, "connection": "closed"}


def generator_function() -> Generator[str, None, None]:
    """Set the connection state to open and close it after the handler returns.

    If an error occurs, set `result` to `"error"`, else set it to `"OK"`.
    """
    try:
        STATE["connection"] = "open"
        yield "hello"
        STATE["result"] = "OK"
    except ValueError:
        STATE["result"] = "error"
    finally:
        STATE["connection"] = "closed"


@get("/{name:str}", dependencies={"message": Provide(generator_function)})
def index(name: str, message: str) -> Dict[str, str]:
    """If `name` is "John", return a message, otherwise raise an error."""
    if name == "John":
        return {name: message}
    raise ValueError()


app = Litestar(route_handlers=[index])



import unittest
from typing import List

import pydantic

from litestar import post
from litestar.testing import create_test_client


class Foo(pydantic.BaseModel):
    bar: str
    baz: List[str]


@post("/")
async def handler(data: Foo) -> Foo:
    return data


class TestApp(unittest.TestCase):
    def test_app(self) -> None:
        assert pydantic.__version__.startswith("1."), pydantic.__version__

        with create_test_client([handler]) as client:
            data = {"bar": "baz", "baz": ["a", "b", "c"]}
            res = client.post("/", json=data)
            self.assertEqual(res.json(), data)



from litestar import Litestar, Request, post
from litestar.middleware.session.server_side import ServerSideSessionConfig
from litestar.testing import AsyncTestClient

session_config = ServerSideSessionConfig()


@post(path="/test", sync_to_thread=False)
def set_session_data(request: Request) -> None:
    request.session["foo"] = "bar"


app = Litestar(route_handlers=[set_session_data], middleware=[session_config.middleware], debug=True)


async def test_set_session_data() -> None:
    async with AsyncTestClient(app=app, session_config=session_config) as client:
        await client.post("/test")
        assert await client.get_session_data() == {"foo": "bar"}



from __future__ import annotations

import msgspec
import pytest
from typing_extensions import Annotated

from litestar import Litestar, get, post
from litestar.datastructures.secret_values import SecretBytes, SecretString
from litestar.openapi.spec.parameter import Parameter as OpenAPIParameter
from litestar.openapi.spec.schema import Schema
from litestar.params import Parameter
from litestar.serialization import default_deserializer, default_serializer
from litestar.testing import create_test_client


def test_secret_string_get_secret() -> None:
    secret_string = SecretString("some_secret_value")
    assert secret_string.get_secret() == "some_secret_value"


def test_secret_string_get_obscured_representation() -> None:
    secret_string = SecretString("some_secret_value")
    assert secret_string.get_obscured() == "******"


def test_secret_string_str() -> None:
    secret_string = SecretString("some_secret_value")
    assert str(secret_string) == "******"


def test_secret_string_repr() -> None:
    secret_string = SecretString("some_secret_value")
    assert repr(secret_string) == "SecretString('******')"


def test_secret_bytes_get_secret() -> None:
    secret_bytes = SecretBytes(b"some_secret_value")
    assert secret_bytes.get_secret() == b"some_secret_value"


def test_secret_bytes_get_obscured_representation() -> None:
    secret_bytes = SecretBytes(b"some_secret_value")
    assert secret_bytes.get_obscured() == b"******"


def test_secret_bytes_str() -> None:
    secret_bytes = SecretBytes(b"some_secret_value")
    assert str(secret_bytes) == str(b"******")


def test_secret_bytes_repr() -> None:
    secret_bytes = SecretBytes(b"some_secret_value")
    assert repr(secret_bytes) == "SecretBytes(b'******')"


def test_secret_string_encode() -> None:
    secret_string = SecretString("some_secret")
    assert default_serializer(secret_string) == "******"


def test_secret_bytes_encode() -> None:
    secret_bytes = SecretBytes(b"some_secret")
    assert default_serializer(secret_bytes) == "******"


def test_secret_string_decode() -> None:
    secret = default_deserializer(SecretString, "super-secret")
    assert isinstance(secret, SecretString)
    assert secret.get_secret() == "super-secret"


def test_secret_bytes_decode() -> None:
    secret = default_deserializer(SecretBytes, b"super-secret")
    assert isinstance(secret, SecretBytes)
    assert secret.get_secret() == b"super-secret"


def test_secret_string_parameter() -> None:
    @get()
    def get_secret(secret: SecretString) -> SecretBytes:
        assert secret.get_secret() == "super-secret"
        return SecretBytes(b"super-secret")

    with create_test_client([get_secret]) as client:
        response = client.get("/?secret=super-secret")
        assert response.status_code == 200
        assert response.json() == "******"


def test_decode_secret_string_on_model() -> None:
    class Model(msgspec.Struct):
        secret: SecretString

    @post(signature_types=[Model])
    async def post_secret(data: Model) -> None:
        assert data.secret.get_secret() == "super"

    with create_test_client([post_secret]) as client:
        response = client.post("/", json={"secret": "super"})
        assert response.status_code == 201


def test_decode_secret_bytes_on_model() -> None:
    class Model(msgspec.Struct):
        secret: SecretBytes

    @post(signature_types=[Model])
    async def post_secret(data: Model) -> None:
        assert data.secret.get_secret() == b"super"

    with create_test_client([post_secret]) as client:
        response = client.post("/", json={"secret": "super"})
        assert response.status_code == 201


@pytest.mark.parametrize(("secret_type",), [(SecretString,), (SecretBytes,)])
def test_decode_secret_string_on_model_client_error(secret_type: type[SecretString | SecretBytes]) -> None:
    model = msgspec.defstruct(name="Model", fields=[("secret", secret_type)])

    @post(signature_namespace={"model": model})
    async def post_secret(data: model) -> None:  # type: ignore[valid-type]
        return None

    with create_test_client([post_secret]) as client:
        response = client.post("/", json={"secret": 123})
        assert response.status_code == 400
        assert response.json() == {
            "status_code": 400,
            "detail": "Validation failed for POST /",
            "extra": [{"message": "Unsupported type: <class 'int'>", "key": "secret", "source": "body"}],
        }


def test_secret_openapi() -> None:
    @get(sync_to_thread=False)
    def get_secret(secret: Annotated[SecretString, Parameter(header="x-secret")]) -> str:
        return secret.get_obscured()

    app = Litestar(route_handlers=[get_secret])
    paths = app.openapi_schema.paths
    assert paths is not None
    op = paths["/"].get
    assert op is not None
    assert op.parameters is not None
    param = op.parameters[0]
    assert isinstance(param, OpenAPIParameter)
    assert param.name == "x-secret"
    assert param.param_in == "header"
    assert isinstance(param.schema, Schema)
    assert param.schema.type == "string"


def test_secret_value_in_model_repr() -> None:
    class Model(msgspec.Struct):
        string: SecretString
        bytes: SecretBytes

    model = Model(string=SecretString("super-secret"), bytes=SecretBytes(b"super-secret"))
    assert repr(model) == "Model(string=SecretString('******'), bytes=SecretBytes(b'******'))"
    assert str(model) == "Model(string=SecretString('******'), bytes=SecretBytes(b'******'))"



from typing import Dict

from litestar import Litestar, get
from litestar.config.app import AppConfig
from litestar.di import Provide
from litestar.plugins import InitPlugin


@get("/", sync_to_thread=False)
def route_handler(name: str) -> Dict[str, str]:
    return {"hello": name}


def get_name() -> str:
    return "world"


class MyPlugin(InitPlugin):
    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        app_config.dependencies["name"] = Provide(get_name, sync_to_thread=False)
        app_config.route_handlers.append(route_handler)
        return app_config


app = Litestar(plugins=[MyPlugin()])

# run: /



"""Plugin for converting exceptions into a problem details response."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Mapping, TypeVar

from typing_extensions import TypeAlias

from litestar.exceptions.http_exceptions import HTTPException
from litestar.plugins.base import InitPlugin
from litestar.response.base import Response

if TYPE_CHECKING:
    from litestar.config.app import AppConfig
    from litestar.connection.request import Request
    from litestar.types.callable_types import ExceptionHandler, ExceptionT

ProblemDetailsExceptionT = TypeVar("ProblemDetailsExceptionT", bound="ProblemDetailsException")
ProblemDetailsExceptionHandlerType: TypeAlias = "Callable[[Request, ProblemDetailsExceptionT], Response]"
ExceptionToProblemDetailMapType: TypeAlias = (
    "Mapping[type[ExceptionT], Callable[[ExceptionT], ProblemDetailsExceptionT]]"
)


def _problem_details_exception_handler(request: Request[Any, Any, Any], exc: ProblemDetailsException) -> Response[Any]:
    return exc.to_response(request)


def _create_exception_handler(
    exc_to_problem_details_exc_fn: Callable[[ExceptionT], ProblemDetailsException], exc_type: type[ExceptionT]
) -> ExceptionHandler[ExceptionT]:
    def _exception_handler(req: Request, exc: exc_type) -> Response:  # type: ignore[valid-type]
        problem_details_exc = exc_to_problem_details_exc_fn(exc)

        return problem_details_exc.to_response(req)

    return _exception_handler


def _http_exception_to_problem_detail_exception(exc: HTTPException) -> ProblemDetailsException:
    return ProblemDetailsException(
        status_code=exc.status_code,
        title=exc.detail,
        extra=exc.extra,
        headers=exc.headers,
    )


class ProblemDetailsException(HTTPException):
    """A problem details exception as per RFC 9457."""

    _PROBLEM_DETAILS_MEDIA_TYPE = "application/problem+json"

    def __init__(
        self,
        *args: Any,
        detail: str = "",
        status_code: int | None = None,
        headers: dict[str, str] | None = None,
        extra: dict[str, Any] | list[Any] | None = None,
        type_: str | None = None,
        title: str | None = None,
        instance: str | None = None,
    ) -> None:
        """Initialize ``ProblemDetailsException``.

        Args:
            *args: if ``detail`` kwarg not provided, first arg should be error detail.
            detail: Exception details or message. Will default to args[0] if not provided.
            status_code: Exception HTTP status code.
            headers: Headers to set on the response.
            extra: An extra mapping to attach to the exception.
            type_: The type field in the problem details.
            title: The title field in the problem details.
            instance: The instance field in the problem details.
        """

        super().__init__(
            *args,
            detail=detail,
            status_code=status_code,
            headers=headers,
            extra=extra,
        )

        self.type_ = type_
        self.title = title
        self.instance = instance

    def to_response(self, request: Request[Any, Any, Any]) -> Response[dict[str, Any]]:
        """Convert the problem details exception into a ``Response.``"""

        problem_details: dict[str, Any] = {"status": self.status_code}
        if self.type_ is not None:
            problem_details["type"] = self.type_
        if self.title is not None:
            problem_details["title"] = self.title
        if self.instance is not None:
            problem_details["instance"] = self.instance
        if self.detail is not None:
            problem_details["detail"] = self.detail

        if extra := self.extra:
            if isinstance(extra, Mapping):
                problem_details.update(extra)
            else:
                problem_details["extra"] = extra

        return Response(
            problem_details,
            headers=self.headers,
            media_type=self._PROBLEM_DETAILS_MEDIA_TYPE,
            status_code=self.status_code,
        )


@dataclass
class ProblemDetailsConfig:
    """The configuration object for ``ProblemDetailsPlugin.``"""

    exception_handler: ProblemDetailsExceptionHandlerType = _problem_details_exception_handler
    """The exception handler used for ``ProblemdetailsException.``"""

    enable_for_all_http_exceptions: bool = False
    """Flag indicating whether to convert all :exc:`HTTPException` into ``ProblemDetailsException.``"""

    exception_to_problem_detail_map: ExceptionToProblemDetailMapType = field(default_factory=dict)
    """A mapping to convert exceptions into ``ProblemDetailsException.``

    All exceptions provided in this will get a custom exception handler where these exceptions
    are converted into ``ProblemDetailException`` before handling them. This can be used to override
    the handler for ``HTTPException`` as well.
    """


class ProblemDetailsPlugin(InitPlugin):
    """A plugin to convert exceptions into problem details as per RFC 9457."""

    def __init__(self, config: ProblemDetailsConfig | None = None):
        self.config = config or ProblemDetailsConfig()

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        app_config.exception_handlers[ProblemDetailsException] = self.config.exception_handler

        if self.config.enable_for_all_http_exceptions:
            app_config.exception_handlers[HTTPException] = _create_exception_handler(
                _http_exception_to_problem_detail_exception, HTTPException
            )

        for exc_type, conversion_fn in self.config.exception_to_problem_detail_map.items():
            app_config.exception_handlers[exc_type] = _create_exception_handler(conversion_fn, exc_type)

        return app_config



import asyncio
from typing import TYPE_CHECKING, List

from home.piccolo_app import APP_CONFIG
from home.tables import Task
from piccolo.apps.user.tables import BaseUser
from piccolo_admin.endpoints import create_admin  # pyright: ignore
from piccolo_api.session_auth.tables import SessionsBase

from litestar import Litestar, asgi, delete, get, patch, post
from litestar.contrib.piccolo_orm import PiccoloORMPlugin
from litestar.exceptions import NotFoundException

if TYPE_CHECKING:
    from litestar.types import Receive, Scope, Send


@asgi("/admin/", is_mount=True)
async def admin(scope: "Scope", receive: "Receive", send: "Send") -> None:
    await create_admin(tables=APP_CONFIG.table_classes)(scope, receive, send)


@get("/tasks", tags=["Task"])
async def tasks() -> List[Task]:
    return await Task.select().order_by(Task.id, ascending=False)


@post("/tasks", tags=["Task"])
async def create_task(data: Task) -> Task:
    task = Task(**data.to_dict())
    await task.save()
    return task


@patch("/tasks/{task_id:int}", tags=["Task"])
async def update_task(task_id: int, data: Task) -> Task:
    task = await Task.objects().get(Task.id == task_id)
    if not task:
        raise NotFoundException("task does not exist")
    for key, value in data.to_dict().items():
        task.id = task_id
        setattr(task, key, value)

    await task.save()
    return task


@delete("/tasks/{task_id:int}", tags=["Task"])
async def delete_task(task_id: int) -> None:
    task = await Task.objects().get(Task.id == task_id)
    if task:
        await task.remove()


async def main() -> None:
    # Creating tables
    await BaseUser.create_table(if_not_exists=True)
    await SessionsBase.create_table(if_not_exists=True)
    await Task.create_table(if_not_exists=True)

    # Creating admin user
    if not await BaseUser.exists().where(BaseUser.email == "admin@test.com"):
        user = BaseUser(
            username="piccolo",
            password="piccolo123",
            email="admin@test.com",
            admin=True,
            active=True,
            superuser=True,
        )
        await user.save()


app = Litestar(
    route_handlers=[
        admin,
        tasks,
        create_task,
        update_task,
        delete_task,
    ],
    plugins=[PiccoloORMPlugin()],
)

if __name__ == "__main__":
    asyncio.run(main())

    import uvicorn

    uvicorn.run(app)



from __future__ import annotations

from typing import Dict, Generic, List, TypeVar

import msgspec
import pytest

from litestar._openapi.datastructures import SchemaRegistry, _get_normalized_schema_key
from litestar.exceptions import ImproperlyConfiguredException
from litestar.openapi.spec import Reference, Schema
from litestar.params import KwargDefinition
from litestar.typing import FieldDefinition
from tests.models import DataclassPerson


@pytest.fixture()
def schema_registry() -> SchemaRegistry:
    return SchemaRegistry()


def test_get_schema_for_field_definition(schema_registry: SchemaRegistry) -> None:
    assert not schema_registry._schema_key_map
    assert not schema_registry._schema_reference_map
    assert not schema_registry._model_name_groups
    field = FieldDefinition.from_annotation(str)
    schema = schema_registry.get_schema_for_field_definition(field)
    key = _get_normalized_schema_key(field)
    assert isinstance(schema, Schema)
    assert key in schema_registry._schema_key_map
    assert not schema_registry._schema_reference_map
    assert len(schema_registry._model_name_groups[key[-1]]) == 1
    assert schema_registry._model_name_groups[key[-1]][0].schema is schema
    assert schema_registry.get_schema_for_field_definition(field) is schema


def test_get_reference_for_field_definition(schema_registry: SchemaRegistry) -> None:
    assert not schema_registry._schema_key_map
    assert not schema_registry._schema_reference_map
    assert not schema_registry._model_name_groups
    field = FieldDefinition.from_annotation(str)
    key = _get_normalized_schema_key(field)

    assert schema_registry.get_reference_for_field_definition(field) is None
    schema_registry.get_schema_for_field_definition(field)
    reference = schema_registry.get_reference_for_field_definition(field)
    assert isinstance(reference, Reference)
    assert id(reference) in schema_registry._schema_reference_map
    assert reference in schema_registry._schema_key_map[key].references


def test_get_normalized_schema_key() -> None:
    class LocalClass(msgspec.Struct):
        id: str

    T = TypeVar("T")

    # replace each of the long strings with underscores with a tuple of strings split at each underscore
    assert _get_normalized_schema_key(FieldDefinition.from_annotation(LocalClass)) == (
        "tests",
        "unit",
        "test_openapi",
        "test_datastructures",
        "test_get_normalized_schema_key.LocalClass",
    )

    assert _get_normalized_schema_key(FieldDefinition.from_annotation(DataclassPerson)) == (
        "tests",
        "models",
        "DataclassPerson",
    )

    builtin_dict = Dict[str, List[int]]
    assert _get_normalized_schema_key(FieldDefinition.from_annotation(builtin_dict)) == (
        "typing",
        "Dict_str_typing.List_int_",
    )

    builtin_with_custom = Dict[str, DataclassPerson]
    assert _get_normalized_schema_key(FieldDefinition.from_annotation(builtin_with_custom)) == (
        "typing",
        "Dict_str_tests.models.DataclassPerson_",
    )

    class LocalGeneric(Generic[T]):
        pass

    assert _get_normalized_schema_key(FieldDefinition.from_annotation(LocalGeneric)) == (
        "tests",
        "unit",
        "test_openapi",
        "test_datastructures",
        "test_get_normalized_schema_key.LocalGeneric",
    )

    generic_int = LocalGeneric[int]
    generic_str = LocalGeneric[str]

    assert _get_normalized_schema_key(FieldDefinition.from_annotation(generic_int)) == (
        "tests",
        "unit",
        "test_openapi",
        "test_datastructures",
        "test_get_normalized_schema_key.LocalGeneric_int_",
    )

    assert _get_normalized_schema_key(FieldDefinition.from_annotation(generic_str)) == (
        "tests",
        "unit",
        "test_openapi",
        "test_datastructures",
        "test_get_normalized_schema_key.LocalGeneric_str_",
    )

    assert _get_normalized_schema_key(FieldDefinition.from_annotation(generic_int)) != _get_normalized_schema_key(
        FieldDefinition.from_annotation(generic_str)
    )


def test_raise_on_override_for_same_field_definition() -> None:
    registry = SchemaRegistry()
    schema = registry.get_schema_for_field_definition(
        FieldDefinition.from_annotation(str, kwarg_definition=KwargDefinition(schema_component_key="foo"))
    )
    # registering the same thing again with the same name should work
    assert (
        registry.get_schema_for_field_definition(
            FieldDefinition.from_annotation(str, kwarg_definition=KwargDefinition(schema_component_key="foo"))
        )
        is schema
    )
    # registering the same *type* with a different name should result in a different schema
    assert (
        registry.get_schema_for_field_definition(
            FieldDefinition.from_annotation(str, kwarg_definition=KwargDefinition(schema_component_key="bar"))
        )
        is not schema
    )
    # registering a different type with a previously used name should raise an exception
    with pytest.raises(ImproperlyConfiguredException):
        registry.get_schema_for_field_definition(
            FieldDefinition.from_annotation(int, kwarg_definition=KwargDefinition(schema_component_key="foo"))
        )



from __future__ import annotations

import asyncio
import dataclasses
from typing import TYPE_CHECKING, AsyncGenerator, Dict, Generator
from unittest.mock import MagicMock

import pytest

from litestar import Controller, Litestar, WebSocket
from litestar.dto import DataclassDTO, dto_field
from litestar.exceptions import ImproperlyConfiguredException
from litestar.handlers.websocket_handlers import websocket_stream
from litestar.testing import create_test_client

if TYPE_CHECKING:
    from litestar.connection import ASGIConnection
    from litestar.handlers.base import BaseRouteHandler


def test_websocket_stream() -> None:
    @websocket_stream("/")
    async def handler(socket: WebSocket) -> AsyncGenerator[str, None]:
        yield "foo"
        yield "bar"

    with create_test_client([handler]) as client, client.websocket_connect("/") as ws:
        assert ws.receive_text(timeout=0.1) == "foo"
        assert ws.receive_text(timeout=0.1) == "bar"


def test_websocket_stream_in_controller() -> None:
    class MyController(Controller):
        @websocket_stream("/")
        async def handler(self, socket: WebSocket) -> AsyncGenerator[str, None]:
            yield "foo"

    with create_test_client([MyController]) as client, client.websocket_connect("/") as ws:
        assert ws.receive_text(timeout=0.1) == "foo"


def test_websocket_stream_without_socket() -> None:
    @websocket_stream("/")
    async def handler() -> AsyncGenerator[str, None]:
        yield "foo"

    with create_test_client([handler]) as client, client.websocket_connect("/") as ws:
        assert ws.receive_text(timeout=0.1) == "foo"


def test_websocket_stream_dependency_injection() -> None:
    async def provide_hello() -> str:
        return "hello"

    # ensure we can inject dependencies
    @websocket_stream("/1", dependencies={"greeting": provide_hello})
    async def handler_one(greeting: str) -> AsyncGenerator[str, None]:
        yield greeting

    # ensure dependency injection also works with 'socket' present
    @websocket_stream("/2", dependencies={"greeting": provide_hello})
    async def handler_two(socket: WebSocket, greeting: str) -> AsyncGenerator[str, None]:
        yield greeting

    with create_test_client([handler_one, handler_two]) as client:
        with client.websocket_connect("/1") as ws:
            assert ws.receive_text(timeout=0.1) == "hello"

        with client.websocket_connect("/2") as ws:
            assert ws.receive_text(timeout=0.1) == "hello"


def test_websocket_stream_dependencies_cleaned_up_after_stream_close() -> None:
    mock = MagicMock()

    async def dep() -> AsyncGenerator[str, None]:
        yield "foo"
        mock()

    @websocket_stream(
        "/",
        dependencies={"message": dep},
        listen_for_disconnect=False,
    )
    async def handler(socket: WebSocket, message: str) -> AsyncGenerator[str, None]:
        yield "one"
        await socket.receive_text()
        yield message

    with create_test_client([handler]) as client, client.websocket_connect("/") as ws:
        assert ws.receive_text(timeout=0.1) == "one"
        assert mock.call_count == 0
        ws.send_text("")
        assert ws.receive_text(timeout=0.1) == "foo"

    assert mock.call_count == 1


def test_websocket_stream_handle_disconnect() -> None:
    @websocket_stream("/")
    async def handler() -> AsyncGenerator[str, None]:
        while True:
            yield "foo"
            # sleep for longer than our read-timeout to ensure we're disconnecting prematurely
            await asyncio.sleep(1)

    with create_test_client([handler]) as client, client.websocket_connect("/") as ws:
        assert ws.receive_text(timeout=0.1) == "foo"

    with create_test_client([handler]) as client, client.websocket_connect("/") as ws:
        # ensure we still disconnect even after receiving some data
        ws.send_text("")
        assert ws.receive_text(timeout=0.1) == "foo"


def test_websocket_stream_send_json() -> None:
    @websocket_stream("/")
    async def handler() -> AsyncGenerator[Dict[str, str], None]:  # noqa: UP006
        yield {"hello": "there"}
        yield {"and": "goodbye"}

    with create_test_client([handler]) as client, client.websocket_connect("/") as ws:
        assert ws.receive_json(timeout=0.1) == {"hello": "there"}
        assert ws.receive_json(timeout=0.1) == {"and": "goodbye"}


def test_websocket_stream_send_json_with_dto() -> None:
    @dataclasses.dataclass
    class Event:
        id: int = dataclasses.field(metadata=dto_field("private"))
        content: str

    @websocket_stream("/", return_dto=DataclassDTO[Event])
    async def handler() -> AsyncGenerator[Event, None]:
        yield Event(id=1, content="hello")

    with create_test_client([handler], signature_types=[Event]) as client, client.websocket_connect("/") as ws:
        assert ws.receive_json(timeout=0.1) == {"content": "hello"}


def test_raises_if_stream_fn_does_not_return_async_generator() -> None:
    with pytest.raises(ImproperlyConfiguredException):

        @websocket_stream("/")  # type: ignore[arg-type]
        def foo() -> Generator[bytes, None, None]:
            yield b""

        Litestar([foo])

    with pytest.raises(ImproperlyConfiguredException):

        @websocket_stream("/")  # type: ignore[arg-type]
        def foo() -> bytes:
            return b""

        Litestar([foo])


def test_websocket_stream_applies_guards() -> None:
    guard_called = False

    async def custom_guard(connection: ASGIConnection, _: BaseRouteHandler) -> None:
        nonlocal guard_called
        guard_called = True

    @websocket_stream("/", guards=[custom_guard])
    async def handler() -> AsyncGenerator[Dict[str, str], None]:  # noqa: UP006
        yield {"Urfaust": "Gespinnst"}
        yield {"Des": "Verderbens"}

    with create_test_client([handler]) as client, client.websocket_connect("/") as ws:
        assert ws.receive_json(timeout=0.1) == {"Urfaust": "Gespinnst"}
        assert ws.receive_json(timeout=0.1) == {"Des": "Verderbens"}
        assert guard_called is True



from unittest.mock import MagicMock, call

import pytest

from litestar import Controller, Request, Router, get
from litestar.status_codes import HTTP_200_OK
from litestar.testing import create_test_client


@pytest.mark.parametrize("sync", [True, False])
@pytest.mark.parametrize("layer", ["app", "router", "controller", "handler"])
def test_after_response_resolution(layer: str, sync: bool) -> None:
    mock = MagicMock()

    if sync:

        def handler(_: Request) -> None:  # pyright: ignore
            mock(layer)

    else:

        async def handler(_: Request) -> None:  # type: ignore[misc]
            mock(layer)

    class MyController(Controller):
        path = "/controller"
        after_response = handler if layer == "controller" else None

        @get("/", after_response=handler if layer == "handler" else None)
        def my_handler(self) -> None:
            return None

    router = Router(
        path="/router", route_handlers=[MyController], after_response=handler if layer == "router" else None
    )

    with create_test_client(route_handlers=[router], after_response=handler if layer == "app" else None) as client:
        response = client.get("/router/controller/")
        assert response.status_code == HTTP_200_OK
        assert all(c == call(layer) for c in mock.call_args_list)



from __future__ import annotations

from dataclasses import dataclass

from litestar import Controller, Litestar, patch, post, put
from litestar.dto import DataclassDTO, DTOConfig, DTOData


@dataclass
class Person:
    name: str
    age: int
    email: str
    id: int


class ReadDTO(DataclassDTO[Person]):
    config = DTOConfig(exclude={"email"})


class WriteDTO(DataclassDTO[Person]):
    config = DTOConfig(exclude={"id"})


class PatchDTO(DataclassDTO[Person]):
    config = DTOConfig(exclude={"id"}, partial=True)


class PersonController(Controller):
    dto = WriteDTO
    return_dto = ReadDTO

    @post("/person", sync_to_thread=False)
    def create_person(self, data: DTOData[Person]) -> Person:
        # Logic for persisting the person goes here
        return data.create_instance(id=1)

    @put("/person/{person_id:int}", sync_to_thread=False)
    def update_person(self, person_id: int, data: DTOData[Person]) -> Person:
        # Usually the Person would be retrieved from a database
        person = Person(id=person_id, name="John", age=50, email="email_of_john@example.com")
        return data.update_instance(person)

    @patch("/person/{person_id:int}", dto=PatchDTO, sync_to_thread=False)
    def patch_person(self, person_id: int, data: DTOData[Person]) -> Person:
        # Usually the Person would be retrieved from a database
        person = Person(id=person_id, name="John", age=50, email="email_of_john@example.com")
        return data.update_instance(person)


app = Litestar(route_handlers=[PersonController])



from litestar import Litestar, Request, get
from litestar.connection.base import empty_receive, empty_send
from litestar.enums import HttpMethod
from litestar.types import Receive, Scope, Send

KITTEN_NAMES_MAP = {
    HttpMethod.GET: "Whiskers",
}


class CustomRequest(Request):
    """Enrich request with the kitten name."""

    __slots__ = ("kitten_name",)

    def __init__(self, scope: Scope, receive: Receive = empty_receive, send: Send = empty_send) -> None:
        """Initialize CustomRequest class."""
        super().__init__(scope=scope, receive=receive, send=send)
        self.kitten_name = KITTEN_NAMES_MAP.get(scope["method"], "Mittens")


@get(path="/kitten-name", sync_to_thread=False)
def get_kitten_name(request: CustomRequest) -> str:
    """Get kitten name based on the HTTP method."""
    return request.kitten_name


app = Litestar(
    route_handlers=[get_kitten_name],
    request_class=CustomRequest,
    debug=True,
)



"""Test exception handlers defined on layers are called for middleware exceptions."""

from __future__ import annotations

from unittest.mock import MagicMock

from litestar import get
from litestar.connection import Request
from litestar.response import Response
from litestar.testing import create_test_client

exception_handler_mock = MagicMock()


def exception_handler(_: Request, __: type[Exception]) -> Response:
    exception_handler_mock()
    return Response(content="", status_code=500)


@get("/raising", sync_to_thread=False)
def raising() -> None:
    raise RuntimeError("damn")


def test_middleware_send_wrapper_called_on_exception() -> None:
    with create_test_client(
        [raising], exception_handlers={RuntimeError: exception_handler}, openapi_config=None
    ) as client:
        client.get("/raising")
        exception_handler_mock.assert_called_once()



from itertools import islice
from typing import List

from polyfactory.factories.pydantic_factory import ModelFactory
from pydantic import BaseModel

from litestar import Litestar, get
from litestar.pagination import AbstractSyncOffsetPaginator, OffsetPagination


class Person(BaseModel):
    id: str
    name: str


class PersonFactory(ModelFactory[Person]):
    __model__ = Person


# we will implement a paginator - the paginator must implement two methods 'get_total' and 'get_items'
# we would usually use a database for this, but for our case we will "fake" the dataset using a factory.


class PersonOffsetPaginator(AbstractSyncOffsetPaginator[Person]):
    def __init__(self) -> None:
        self.data = PersonFactory.batch(50)

    def get_total(self) -> int:
        return len(self.data)

    def get_items(self, limit: int, offset: int) -> List[Person]:
        return list(islice(islice(self.data, offset, None), limit))


paginator = PersonOffsetPaginator()


# we now create a regular handler. The handler will receive two query parameters - 'limit' and 'offset', which
# we will pass to the paginator.
@get("/people", sync_to_thread=False)
def people_handler(limit: int, offset: int) -> OffsetPagination[Person]:
    return paginator(limit=limit, offset=offset)


app = Litestar(route_handlers=[people_handler])



from __future__ import annotations

from contextlib import asynccontextmanager
from datetime import date, datetime
from typing import AsyncIterator
from uuid import UUID

import anyio
from rich import get_console
from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine
from sqlalchemy.orm import Mapped

from litestar.plugins.sqlalchemy import base, repository

console = get_console()


# the SQLAlchemy base includes a declarative model for you to use in your models.
# The `Base` class includes a `UUID` based primary key (`id`)
class Author(base.UUIDBase):
    name: Mapped[str]
    dob: Mapped[date]
    dod: Mapped[date | None]


class AuthorRepository(repository.SQLAlchemyAsyncRepository[Author]):
    """Author repository."""

    model_type = Author


engine = create_async_engine(
    "sqlite+aiosqlite:///test.sqlite",
    future=True,
)
session_factory = async_sessionmaker(engine, expire_on_commit=False)


# let's make a simple context manager as an example here.
@asynccontextmanager
async def repository_factory() -> AsyncIterator[AuthorRepository]:
    async with session_factory() as db_session:
        try:
            yield AuthorRepository(session=db_session)
        except Exception:  # noqa: BLE001
            await db_session.rollback()
        else:
            await db_session.commit()


async def create_author() -> Author:
    async with repository_factory() as repo:
        obj = await repo.add(
            Author(
                name="F. Scott Fitzgerald",
                dob=datetime.strptime("1896-09-24", "%Y-%m-%d").date(),
            )
        )
        console.print(f"Created Author record for {obj.name} with primary key {obj.id}.")
        return obj


async def update_author(obj: Author) -> Author:
    async with repository_factory() as repo:
        obj = await repo.update(obj)
        console.print(f"Updated Author record for {obj.name} with primary key {obj.id}.")
        return obj


async def remove_author(id: UUID) -> Author:
    async with repository_factory() as repo:
        obj = await repo.delete(id)
        console.print(f"Deleted Author record for {obj.name} with primary key {obj.id}.")
        return obj


async def get_author_if_exists(id: UUID) -> Author | None:
    async with repository_factory() as repo:
        obj = await repo.get_one_or_none(id=id)
        if obj is not None:
            console.print(f"Found Author record for {obj.name} with primary key {obj.id}.")
        else:
            console.print(f"Could not find Author with primary key {id}.")
        return obj


async def run_script() -> None:
    """Load data from a fixture."""
    async with engine.begin() as conn:
        await conn.run_sync(base.UUIDBase.metadata.create_all)

    # 1) create a new Author record.
    console.print("1) Adding a new record")
    author = await create_author()
    author_id = author.id

    # 2) Let's update the Author record.
    console.print("2) Updating a record.")
    author.dod = datetime.strptime("1940-12-21", "%Y-%m-%d").date()
    await update_author(author)

    # 3) Let's delete the record we just created.
    console.print("3) Removing a record.")
    await remove_author(author_id)

    # 4) Let's verify the record no longer exists.
    console.print("4) Select one or none.")
    _should_be_none = await get_author_if_exists(author_id)


if __name__ == "__main__":
    anyio.run(run_script)



import logging
from typing import Dict

from litestar import Litestar, Response, get
from litestar.background_tasks import BackgroundTask

logger = logging.getLogger(__name__)


async def logging_task(identifier: str, message: str) -> None:
    logger.info("%s: %s", identifier, message)


@get("/", sync_to_thread=False)
def greeter(name: str) -> Response[Dict[str, str]]:
    return Response(
        {"hello": name},
        background=BackgroundTask(logging_task, "greeter", message=f"was called with name {name}"),
    )


app = Litestar(route_handlers=[greeter])



from typing import Any, Dict, List, Literal, Mapping, Optional, Union, cast

import pytest

from litestar import get
from litestar.config.cors import CORSConfig
from litestar.middleware._internal.cors import CORSMiddleware
from litestar.status_codes import HTTP_200_OK, HTTP_404_NOT_FOUND
from litestar.testing import create_test_client
from litestar.types.asgi_types import Method


def test_setting_cors_middleware() -> None:
    cors_config = CORSConfig()  # pyright: ignore
    assert cors_config.allow_credentials is False
    assert cors_config.allow_headers == ["*"]
    assert cors_config.allow_methods == ["*"]
    assert cors_config.allow_origins == ["*"]
    assert cors_config.allow_origin_regex is None
    assert cors_config.max_age == 600
    assert cors_config.expose_headers == []

    with create_test_client(cors_config=cors_config) as client:
        unpacked_middleware = []
        cur = client.app.asgi_handler
        while hasattr(cur, "app"):
            unpacked_middleware.append(cur)
            cur = cast("Any", cur.app)
        unpacked_middleware.append(cur)
        assert len(unpacked_middleware) == 4
        cors_middleware = cast("Any", unpacked_middleware[0])
        assert isinstance(cors_middleware, CORSMiddleware)
        assert cors_middleware.config.allow_headers == ["*"]
        assert cors_middleware.config.allow_methods == ["*"]
        assert cors_middleware.config.allow_origins == cors_config.allow_origins
        assert cors_middleware.config.allow_origin_regex == cors_config.allow_origin_regex


@pytest.mark.parametrize("origin", [None, "http://www.example.com", "https://moishe.zuchmir.com"])
@pytest.mark.parametrize("allow_origins", ["*", "http://www.example.com", "https://moishe.zuchmir.com"])
@pytest.mark.parametrize("allow_credentials", [True, False])
@pytest.mark.parametrize(
    "expose_headers", [["x-first-header", "x-second-header", "x-third-header"], ["*"], ["x-first-header"]]
)
@pytest.mark.parametrize(
    "allow_headers", [["x-first-header", "x-second-header", "x-third-header"], ["*"], ["x-first-header"]]
)
@pytest.mark.parametrize("allow_methods", [["GET", "POST", "PUT", "DELETE"], ["GET", "POST"], ["GET"]])
def test_cors_simple_response(
    origin: Optional[str],
    allow_origins: List[str],
    allow_credentials: bool,
    expose_headers: List[str],
    allow_headers: List[str],
    allow_methods: List[Union[Literal["*"], "Method"]],
) -> None:
    @get("/")
    def handler() -> Dict[str, str]:
        return {"hello": "world"}

    cors_config = CORSConfig(
        allow_origins=allow_origins,
        allow_credentials=allow_credentials,
        expose_headers=expose_headers,
        allow_headers=allow_headers,
        allow_methods=allow_methods,
    )

    with create_test_client(handler, cors_config=cors_config) as client:
        headers: Mapping[str, str] = {"Origin": origin} if origin else {}
        response = client.get("/", headers=headers)
        assert response.status_code == HTTP_200_OK
        assert response.json() == {"hello": "world"}
        assert cors_config.expose_headers == expose_headers
        assert cors_config.allow_origins == allow_origins
        assert cors_config.allow_credentials == allow_credentials
        assert cors_config.allow_headers == allow_headers
        assert cors_config.allow_methods == allow_methods

        if origin:
            if cors_config.is_allow_all_origins:
                assert response.headers.get("Access-Control-Allow-Origin") == "*"
            if cors_config.allow_credentials:
                assert response.headers.get("Access-Control-Allow-Credentials") == "true"
            if cors_config.expose_headers:
                assert response.headers.get("Access-Control-Expose-Headers") == ", ".join(
                    sorted(set(cors_config.expose_headers))
                )
            if cors_config.allow_headers:
                assert response.headers.get("Access-Control-Allow-Headers") == ", ".join(
                    sorted(set(cors_config.allow_headers))
                )
            if cors_config.allow_methods:
                assert response.headers.get("Access-Control-Allow-Methods") == ", ".join(
                    sorted(set(cors_config.allow_methods))
                )
        else:
            assert "Access-Control-Allow-Origin" not in response.headers
            assert "Access-Control-Allow-Credentials" not in response.headers
            assert "Access-Control-Expose-Headers" not in response.headers
            assert "Access-Control-Allow-Headers" not in response.headers
            assert "Access-Control-Allow-Methods" not in response.headers


@pytest.mark.parametrize("origin, should_apply_cors", (("http://www.example.com", True), (None, False)))
def test_cors_applied_on_exception_response_if_origin_is_present(
    origin: Optional[str], should_apply_cors: bool
) -> None:
    @get("/")
    def handler() -> Dict[str, str]:
        return {"hello": "world"}

    cors_config = CORSConfig(allow_origins=["http://www.example.com"])

    with create_test_client(handler, cors_config=cors_config) as client:
        headers: Mapping[str, str] = {"Origin": origin} if origin else {}
        response = client.get("/abc", headers=headers)
        assert response.status_code == HTTP_404_NOT_FOUND
        if should_apply_cors:
            assert response.headers.get("Access-Control-Allow-Origin") == origin
        else:
            assert not response.headers.get("Access-Control-Allow-Origin")


@pytest.mark.parametrize(
    "allow_origin,origin,host,should_allow",
    [
        ("httpx://good.example", "https://goodXexample", "example.com", False),
        ("https://*good.example", "https://very.good.example", "very.good.example", True),
        ("https://*good.example", "https://verygood.example", "vergood.example", True),
        ("https://*good.example", "https://good.example", "good.example", True),
        ("https://*good.example", "https://bad.example", "bad.example", False),
        ("https://*.good.example", "https://very.good.example", "very.good.example", True),
        ("https://*.good.example", "https://verygood.example", "verygood.example", False),
        ("https://*.good.example", "https://some.verygood.example", "verygood.example", False),
        ("https://*.good.example", "https://good.example", "good.example", False),
    ],
)
def test_cors_test_regex_escape(allow_origin: str, origin: str, host: str, should_allow: bool) -> None:
    @get("/")
    async def handler() -> None:
        return None

    with create_test_client(
        [handler],
        cors_config=CORSConfig(
            allow_origins=[allow_origin],
            allow_credentials=True,
        ),
    ) as client:
        res = client.get("/", headers={"Origin": origin, "Host": host})

    if should_allow:
        assert "Access-Control-Allow-Origin" in res.headers
    else:
        assert "Access-Control-Allow-Origin" not in res.headers



from dataclasses import dataclass
from typing import List

from litestar import Litestar, post


@dataclass
class TodoItem:
    title: str
    done: bool


TODO_LIST: List[TodoItem] = []


@post("/")
async def add_item(data: TodoItem) -> List[TodoItem]:
    TODO_LIST.append(data)
    return TODO_LIST


app = Litestar([add_item])



from typing import Any, Type

from msgspec import Struct

from litestar import Litestar, post


class TenantUser:
    """Custom Type that represents a user associated to a tenant

    Parsed from / serialized to a combined tenant + user id string of the form

        TENANTPREFIX_USERID

    i.e. separated by underscore.
    """

    tenant_prefix: str
    user_id: str

    def __init__(self, tenant_prefix: str, user_id: str) -> None:
        self.tenant_prefix = tenant_prefix
        self.user_id = user_id

    @classmethod
    def from_string(cls, s: str) -> "TenantUser":
        splits = s.split("_", maxsplit=1)
        if len(splits) < 2:
            raise ValueError(
                "Could not split up tenant user id string. "
                "Expecting underscore for separation of tenant prefix and user id."
            )
        return cls(tenant_prefix=splits[0], user_id=splits[1])

    def to_combined_str(self) -> str:
        return self.tenant_prefix + "_" + self.user_id


def tenant_user_type_predicate(type: Type) -> bool:
    return type is TenantUser


def tenant_user_enc_hook(u: TenantUser) -> Any:
    return u.to_combined_str()


def tenant_user_dec_hook(tenant_user_id_str: str) -> TenantUser:
    return TenantUser.from_string(tenant_user_id_str)


def general_dec_hook(type: Type, obj: Any) -> Any:
    if tenant_user_type_predicate(type):
        return tenant_user_dec_hook(obj)

    raise NotImplementedError(f"Encountered unknown type during decoding: {type!s}")


class UserAsset(Struct):
    user: TenantUser
    name: str


@post("/asset", sync_to_thread=False)
def create_asset(
    data: UserAsset,
) -> UserAsset:
    assert isinstance(data.user, TenantUser)
    return data


app = Litestar(
    [create_asset],
    type_encoders={TenantUser: tenant_user_enc_hook},  # tell litestar how to encode TenantUser
    type_decoders=[(tenant_user_type_predicate, general_dec_hook)],  # tell litestar how to decode TenantUser
)

# run: /asset -X POST -H "Content-Type: application/json" -d '{"name":"SomeAsset","user":"TenantA_Somebody"}'



from litestar import Litestar
from litestar.plugins.prometheus import PrometheusConfig, PrometheusController


def create_app(group_path: bool = False):
    # Default app name and prefix is litestar.
    prometheus_config = PrometheusConfig(group_path=group_path)

    # By default the metrics are available in prometheus format and the path is set to '/metrics'.
    # If you want to change the path and format you can do it by subclassing the PrometheusController class.

    # Creating the litestar app instance with our custom PrometheusConfig and PrometheusController.
    return Litestar(route_handlers=[PrometheusController], middleware=[prometheus_config.middleware])



from litestar import Litestar, MediaType, Request, Response, get
from litestar.exceptions import HTTPException, ValidationException
from litestar.status_codes import HTTP_500_INTERNAL_SERVER_ERROR


def validation_exception_handler(request: Request, exc: ValidationException) -> Response:
    return Response(
        media_type=MediaType.TEXT,
        content=f"validation error: {exc.detail}",
        status_code=400,
    )


def internal_server_error_handler(request: Request, exc: Exception) -> Response:
    return Response(
        media_type=MediaType.TEXT,
        content=f"server error: {exc}",
        status_code=500,
    )


def value_error_handler(request: Request, exc: ValueError) -> Response:
    return Response(
        media_type=MediaType.TEXT,
        content=f"value error: {exc}",
        status_code=400,
    )


@get("/validation-error")
async def validation_error(some_query_param: str) -> str:
    return some_query_param


@get("/server-error")
async def server_error() -> None:
    raise HTTPException()


@get("/value-error")
async def value_error() -> None:
    raise ValueError("this is wrong")


app = Litestar(
    route_handlers=[validation_error, server_error, value_error],
    exception_handlers={
        ValidationException: validation_exception_handler,
        HTTP_500_INTERNAL_SERVER_ERROR: internal_server_error_handler,
        ValueError: value_error_handler,
    },
)


# run: /validation-error
# run: /server-error
# run: /value-error



from litestar.middleware import MiddlewareProtocol
from litestar.types import ASGIApp, Receive, Scope, Send


class MyMiddleware(MiddlewareProtocol):
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        # do stuff
        await self.app(scope, receive, send)



from contextlib import asynccontextmanager
from typing import AsyncGenerator

from sqlalchemy.ext.asyncio import create_async_engine

from litestar import Litestar


@asynccontextmanager
async def db_connection(app: Litestar) -> AsyncGenerator[None, None]:
    engine = getattr(app.state, "engine", None)
    if engine is None:
        engine = create_async_engine("postgresql+asyncpg://postgres:mysecretpassword@pg.db:5432/db")
        app.state.engine = engine

    try:
        yield
    finally:
        await engine.dispose()


app = Litestar(lifespan=[db_connection])



from pathlib import Path

from litestar import Litestar
from litestar.static_files import create_static_files_router

ASSETS_DIR = Path("assets")


def on_startup():
    ASSETS_DIR.mkdir(exist_ok=True)
    ASSETS_DIR.joinpath("hello.txt").write_text("Hello, world!")


app = Litestar(
    route_handlers=[
        create_static_files_router(path="/static", directories=["assets"]),
    ],
    on_startup=[on_startup],
)


#  run: /static/hello.txt



from dataclasses import dataclass

from litestar import Litestar, post
from litestar.plugins.problem_details import ProblemDetailsConfig, ProblemDetailsException, ProblemDetailsPlugin


@dataclass
class PurchaseItem:
    item_id: int
    quantity: int


@post("/purchase")
async def purchase(data: PurchaseItem) -> None:
    # Logic to check if the user has enough credit to buy the item.
    # We assume the user does not have enough credit.

    raise ProblemDetailsException(
        type_="https://example.com/probs/out-of-credit",
        title="You do not have enough credit.",
        detail="Your current balance is 30, but that costs 50.",
        instance="/account/12345/msgs/abc",
        extra={"balance": 30},
    )


problem_details_plugin = ProblemDetailsPlugin(ProblemDetailsConfig())
app = Litestar(route_handlers=[purchase], plugins=[problem_details_plugin])

# run: /purchase --header "Content-Type: application/json" --request POST --data '{"item_id": 1234, "quantity": 2}'



from __future__ import annotations

from unittest.mock import ANY

from litestar.status_codes import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT
from litestar.testing import TestClient


def test_create_user(user_data: dict) -> None:
    from docs.examples.data_transfer_objects.defining_dtos_on_layers import app

    with TestClient(app=app) as client:
        response = client.post("/", json=user_data)

    assert response.status_code == HTTP_201_CREATED
    assert response.json() == {"id": ANY, "name": "Mr Sunglass", "email": "mr.sunglass@example.com", "age": 30}


def test_get_users() -> None:
    from docs.examples.data_transfer_objects.defining_dtos_on_layers import app

    with TestClient(app=app) as client:
        response = client.get("/")

    assert response.status_code == HTTP_200_OK
    assert response.json() == [{"id": ANY, "name": "Mr Sunglass", "email": "mr.sunglass@example.com", "age": 30}]


def test_get_user() -> None:
    from docs.examples.data_transfer_objects.defining_dtos_on_layers import app

    with TestClient(app=app) as client:
        response = client.get("/a3cad591-5b01-4341-ae8f-94f78f790674")

    assert response.status_code == HTTP_200_OK
    assert response.json() == {
        "id": "a3cad591-5b01-4341-ae8f-94f78f790674",
        "name": "Mr Sunglass",
        "email": "mr.sunglass@example.com",
        "age": 30,
    }


def test_update_user(user_data: dict) -> None:
    from docs.examples.data_transfer_objects.defining_dtos_on_layers import app

    with TestClient(app=app) as client:
        response = client.put("/a3cad591-5b01-4341-ae8f-94f78f790674", json=user_data)

    assert response.status_code == HTTP_200_OK
    assert response.json() == {"id": ANY, "name": "Mr Sunglass", "email": "mr.sunglass@example.com", "age": 30}


def test_delete_user() -> None:
    from docs.examples.data_transfer_objects.defining_dtos_on_layers import app

    with TestClient(app=app) as client:
        response = client.delete("/a3cad591-5b01-4341-ae8f-94f78f790674")

    assert response.status_code == HTTP_204_NO_CONTENT
    assert response.content == b""



from litestar import Litestar, MediaType, get
from litestar.middleware.rate_limit import RateLimitConfig

rate_limit_config = RateLimitConfig(rate_limit=("minute", 1), exclude=["/schema"])


@get("/", media_type=MediaType.TEXT, sync_to_thread=False)
def handler() -> str:
    """Handler which should not be accessed more than once per minute."""
    return "ok"


app = Litestar(route_handlers=[handler], middleware=[rate_limit_config.middleware])



from __future__ import annotations

from inspect import getmro
from sys import exc_info
from traceback import format_exception
from typing import TYPE_CHECKING, Any, Type, Union, cast

from litestar.enums import ScopeType
from litestar.exceptions import HTTPException, LitestarException, WebSocketException
from litestar.exceptions.responses import create_exception_response
from litestar.exceptions.responses._debug_response import (
    create_debug_response,
)
from litestar.status_codes import HTTP_500_INTERNAL_SERVER_ERROR
from litestar.utils.deprecation import warn_deprecation
from litestar.utils.empty import value_or_raise
from litestar.utils.scope.state import ScopeState

if TYPE_CHECKING:
    from starlette.exceptions import HTTPException as StarletteHTTPException

    from litestar import Response
    from litestar.app import Litestar
    from litestar.connection import Request
    from litestar.handlers import BaseRouteHandler
    from litestar.logging import BaseLoggingConfig
    from litestar.types import (
        ASGIApp,
        ExceptionHandler,
        ExceptionHandlersMap,
        Logger,
        Message,
        Receive,
        Scope,
        Send,
    )
    from litestar.types.asgi_types import WebSocketCloseEvent

__all__ = ("ExceptionHandlerMiddleware",)


def get_exception_handler(exception_handlers: ExceptionHandlersMap, exc: Exception) -> ExceptionHandler | None:
    """Given a dictionary that maps exceptions and status codes to handler functions, and an exception, returns the
    appropriate handler if existing.

    Status codes are given preference over exception type.

    If no status code match exists, each class in the MRO of the exception type is checked and
    the first matching handler is returned.

    Finally, if a ``500`` handler is registered, it will be returned for any exception that isn't a
    subclass of :class:`HTTPException <litestar.exceptions.HTTPException>`.

    Args:
        exception_handlers: Mapping of status codes and exception types to handlers.
        exc: Exception Instance to be resolved to a handler.

    Returns:
        Optional exception handler callable.
    """
    if not exception_handlers:
        return None

    default_handler: ExceptionHandler | None = None
    if isinstance(exc, HTTPException):
        if exception_handler := exception_handlers.get(exc.status_code):
            return exception_handler
    else:
        default_handler = exception_handlers.get(HTTP_500_INTERNAL_SERVER_ERROR)

    return next(
        (exception_handlers[cast("Type[Exception]", cls)] for cls in getmro(type(exc)) if cls in exception_handlers),
        default_handler,
    )


def _starlette_exception_handler(request: Request[Any, Any, Any], exc: StarletteHTTPException) -> Response:
    return create_exception_response(
        request=request,
        exc=HTTPException(
            detail=exc.detail,
            status_code=exc.status_code,
            headers=exc.headers,  # type: ignore[arg-type]
        ),
    )


class ExceptionHandlerMiddleware:
    """Middleware used to wrap an ASGIApp inside a try catch block and handle any exceptions raised.

    This used in multiple layers of Litestar.
    """

    def __init__(
        self, app: ASGIApp, debug: bool | None, exception_handlers: ExceptionHandlersMap | None = None
    ) -> None:
        """Initialize ``ExceptionHandlerMiddleware``.

        Args:
            app: The ``next`` ASGI app to call.
            debug: Whether ``debug`` mode is enabled. Deprecated. Debug mode will be inferred from the request scope
            exception_handlers: A dictionary mapping status codes and/or exception types to handler functions.

        .. deprecated:: 2.0.0
            The ``debug`` parameter is deprecated. It will be inferred from the request scope

        .. deprecated:: 2.9.0
            The ``exception_handlers`` parameter is deprecated. It will be inferred from the application or the
            route handler.
        """
        self.app = app
        self.exception_handlers = exception_handlers
        self.debug = debug

        if debug is not None:
            warn_deprecation(
                "2.0.0",
                deprecated_name="debug",
                kind="parameter",
                info="Debug mode will be inferred from the request scope",
                removal_in="3.0.0",
            )

        if exception_handlers is not None:
            warn_deprecation(
                "2.9.0",
                deprecated_name="exception_handlers",
                kind="parameter",
                info="It will be inferred from the application or the route handler",
                removal_in="3.0.0",
            )

        self._get_debug = self._get_debug_scope if debug is None else lambda *a: debug

    @staticmethod
    def _get_debug_scope(scope: Scope) -> bool:
        return scope["litestar_app"].debug

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """ASGI-callable.

        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """
        scope_state = ScopeState.from_scope(scope)

        async def capture_response_started(event: Message) -> None:
            if event["type"] == "http.response.start":
                scope_state.response_started = True
            await send(event)

        try:
            await self.app(scope, receive, capture_response_started)
        except Exception as e:
            if scope_state.response_started:
                raise LitestarException("Exception caught after response started") from e

            litestar_app = scope["litestar_app"]

            if litestar_app.logging_config and (logger := litestar_app.logger):
                self.handle_exception_logging(logger=logger, logging_config=litestar_app.logging_config, scope=scope)

            for hook in litestar_app.after_exception:
                await hook(e, scope)

            if litestar_app.pdb_on_exception:
                litestar_app.debugger_module.post_mortem()

            if scope["type"] == ScopeType.HTTP:
                await self.handle_request_exception(
                    litestar_app=litestar_app, scope=scope, receive=receive, send=send, exc=e
                )
            else:
                await self.handle_websocket_exception(send=send, exc=e)

    async def handle_request_exception(
        self, litestar_app: Litestar, scope: Scope, receive: Receive, send: Send, exc: Exception
    ) -> None:
        """Handle exception raised inside 'http' scope routes.

        Args:
            litestar_app: The litestar app instance.
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.
            exc: The caught exception.

        Returns:
            None.
        """

        exception_handlers = (
            value_or_raise(ScopeState.from_scope(scope).exception_handlers)
            if self.exception_handlers is None
            else self.exception_handlers
        )
        exception_handler = get_exception_handler(exception_handlers, exc) or self.default_http_exception_handler
        request: Request[Any, Any, Any] = litestar_app.request_class(scope=scope, receive=receive, send=send)
        response = exception_handler(request, exc)
        route_handler: BaseRouteHandler | None = scope.get("route_handler")
        type_encoders = route_handler.resolve_type_encoders() if route_handler else litestar_app.type_encoders
        await response.to_asgi_response(app=None, request=request, type_encoders=type_encoders)(
            scope=scope, receive=receive, send=send
        )

    @staticmethod
    async def handle_websocket_exception(send: Send, exc: Exception) -> None:
        """Handle exception raised inside 'websocket' scope routes.

        Args:
            send: The ASGI send function.
            exc: The caught exception.

        Returns:
            None.
        """
        code = 4000 + HTTP_500_INTERNAL_SERVER_ERROR
        reason = "Internal Server Error"
        if isinstance(exc, WebSocketException):
            code = exc.code
            reason = exc.detail
        elif isinstance(exc, LitestarException):
            reason = exc.detail

        event: WebSocketCloseEvent = {"type": "websocket.close", "code": code, "reason": reason}
        await send(event)

    def default_http_exception_handler(self, request: Request, exc: Exception) -> Response[Any]:
        """Handle an HTTP exception by returning the appropriate response.

        Args:
            request: An HTTP Request instance.
            exc: The caught exception.

        Returns:
            An HTTP response.
        """
        status_code = exc.status_code if isinstance(exc, HTTPException) else HTTP_500_INTERNAL_SERVER_ERROR
        if status_code == HTTP_500_INTERNAL_SERVER_ERROR and self._get_debug_scope(request.scope):
            return create_debug_response(request=request, exc=exc)
        return create_exception_response(request=request, exc=exc)

    def handle_exception_logging(self, logger: Logger, logging_config: BaseLoggingConfig, scope: Scope) -> None:
        """Handle logging - if the litestar app has a logging config in place.

        Args:
            logger: A logger instance.
            logging_config: Logging Config instance.
            scope: The ASGI connection scope.

        Returns:
            None
        """
        exc = exc_info()
        exc_detail: set[Union[Exception, int]] = {exc[0], getattr(exc[0], "status_code", None)}  # type: ignore[arg-type]  # noqa: UP007

        if (
            (
                logging_config.log_exceptions == "always"
                or (logging_config.log_exceptions == "debug" and self._get_debug_scope(scope))
            )
            and logging_config.exception_logging_handler
            and exc_detail.isdisjoint(logging_config.disable_stack_trace)
        ):
            logging_config.exception_logging_handler(logger, scope, format_exception(*exc))



# ruff: noqa: F401
from __future__ import annotations

import importlib
import sys
from importlib.util import cache_from_source
from pathlib import Path

import pytest


def purge_module(module_names: list[str], path: str | Path) -> None:
    for name in module_names:
        if name in sys.modules:
            del sys.modules[name]
    Path(cache_from_source(str(path))).unlink(missing_ok=True)


def test_deprecated_prometheus_imports() -> None:
    purge_module(["litestar.contrib.prometheus"], __file__)
    with pytest.warns(
        DeprecationWarning, match="importing PrometheusMiddleware from 'litestar.contrib.prometheus' is deprecated"
    ):
        from litestar.contrib.prometheus import PrometheusMiddleware

    purge_module(["litestar.contrib.prometheus"], __file__)
    with pytest.warns(
        DeprecationWarning, match="importing PrometheusConfig from 'litestar.contrib.prometheus' is deprecated"
    ):
        from litestar.contrib.prometheus import PrometheusConfig

    purge_module(["litestar.contrib.prometheus"], __file__)
    with pytest.warns(
        DeprecationWarning, match="importing PrometheusController from 'litestar.contrib.prometheus' is deprecated"
    ):
        from litestar.contrib.prometheus import PrometheusController


def test_deprecated_prometheus_middleware_imports() -> None:
    purge_module(["litestar.contrib.prometheus.middleware"], __file__)
    with pytest.warns(
        DeprecationWarning,
        match="importing PrometheusMiddleware from 'litestar.contrib.prometheus.middleware' is deprecated",
    ):
        from litestar.contrib.prometheus.middleware import PrometheusMiddleware


def test_deprecated_prometheus_config_imports() -> None:
    purge_module(["litestar.contrib.prometheus.config"], __file__)
    with pytest.warns(
        DeprecationWarning,
        match="importing PrometheusConfig from 'litestar.contrib.prometheus.config' is deprecated",
    ):
        from litestar.contrib.prometheus.config import PrometheusConfig


def test_deprecated_prometheus_controller_imports() -> None:
    purge_module(["litestar.contrib.prometheus.controller"], __file__)
    with pytest.warns(
        DeprecationWarning,
        match="importing PrometheusController from 'litestar.contrib.prometheus.controller' is deprecated",
    ):
        from litestar.contrib.prometheus.controller import PrometheusController



from __future__ import annotations

from typing import Dict
from unittest.mock import MagicMock

from litestar import Controller, Litestar, Router, post
from litestar.dto import AbstractDTO, DTOConfig
from litestar.dto._backend import DTOBackend
from litestar.testing import create_test_client

from . import Model


def test_dto_defined_on_handler(ModelDataDTO: type[AbstractDTO]) -> None:
    @post(dto=ModelDataDTO, signature_types=[Model])
    def handler(data: Model) -> Model:
        assert data == Model(a=1, b="2")
        return data

    with create_test_client(route_handlers=handler) as client:
        response = client.post("/", json={"what": "ever"})
        assert response.status_code == 201
        assert response.json() == {"a": 1, "b": "2"}


def test_dto_defined_on_controller(ModelDataDTO: type[AbstractDTO]) -> None:
    class MyController(Controller):
        dto = ModelDataDTO

        @post()
        def handler(self, data: Model) -> Model:
            assert data == Model(a=1, b="2")
            return data

    with create_test_client(route_handlers=MyController) as client:
        response = client.post("/", json={"what": "ever"})
        assert response.status_code == 201
        assert response.json() == {"a": 1, "b": "2"}


def test_dto_defined_on_router(ModelDataDTO: type[AbstractDTO]) -> None:
    @post()
    def handler(data: Model) -> Model:
        assert data == Model(a=1, b="2")
        return data

    router = Router(path="/", route_handlers=[handler], dto=ModelDataDTO)

    with create_test_client(route_handlers=router) as client:
        response = client.post("/", json={"what": "ever"})
        assert response.status_code == 201
        assert response.json() == {"a": 1, "b": "2"}


def test_dto_defined_on_app(ModelDataDTO: type[AbstractDTO]) -> None:
    @post()
    def handler(data: Model) -> Model:
        assert data == Model(a=1, b="2")
        return data

    with create_test_client(route_handlers=handler, dto=ModelDataDTO) as client:
        response = client.post("/", json={"what": "ever"})
        assert response.status_code == 201
        assert response.json() == {"a": 1, "b": "2"}


def test_set_dto_none_disables_inherited_dto(ModelDataDTO: type[AbstractDTO]) -> None:
    @post(dto=None, signature_namespace={"dict": Dict})
    def handler(data: dict[str, str]) -> dict[str, str]:
        assert data == {"hello": "world"}
        return data

    mock_dto = MagicMock(spec=ModelDataDTO)

    with create_test_client(route_handlers=handler, dto=mock_dto) as client:  # pyright:ignore
        response = client.post("/", json={"hello": "world"})
        assert response.status_code == 201
        assert response.json() == {"hello": "world"}
        mock_dto.assert_not_called()


def test_dto_and_return_dto(ModelDataDTO: type[AbstractDTO], ModelReturnDTO: type[AbstractDTO]) -> None:
    @post()
    def handler(data: Model) -> Model:
        assert data == Model(a=1, b="2")
        return data

    with create_test_client(route_handlers=handler, dto=ModelDataDTO, return_dto=ModelReturnDTO) as client:
        response = client.post("/", json={"what": "ever"})
        assert response.status_code == 201
        assert response.json() == {"a": 1, "b": "2"}


def test_enable_experimental_backend_override_in_dto_config(ModelDataDTO: type[AbstractDTO]) -> None:
    ModelDataDTO.config = DTOConfig(experimental_codegen_backend=False)

    @post(dto=ModelDataDTO, signature_types=[Model])
    def handler(data: Model) -> Model:
        return data

    Litestar(route_handlers=[handler])

    backend = handler.resolve_data_dto()._dto_backends[handler.handler_id]["data_backend"]  # type: ignore[union-attr]
    assert isinstance(backend, DTOBackend)


def test_use_codegen_backend_by_default(ModelDataDTO: type[AbstractDTO]) -> None:
    ModelDataDTO.config = DTOConfig()

    @post(dto=ModelDataDTO, signature_types=[Model])
    def handler(data: Model) -> Model:
        return data

    Litestar(route_handlers=[handler])

    backend = handler.resolve_data_dto()._dto_backends[handler.handler_id]["data_backend"]  # type: ignore[union-attr]
    assert isinstance(backend, DTOBackend)
